"""A trajectory generator that return signals for alternating legs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import attr
import gin
from gym import spaces
import numpy as np
from pybullet_envs.minitaur.envs.utilities import laikago_pose_utils

TROT_GAIT = "trot"
PACE_GAIT = "pace"
NUM_MOTORS_LAIKAGO = 12
STD_FOR_GAUSSIAN_TRAJECTORY = 0.15
MOTION_FREQUENCY = 1.0
MOTION_AMPLITUDE = 0.25
ACTION_BOUND = 0.25


# TODO(b/131193449): Add a test to this class.
@gin.configurable
class LaikagoAlternatingLegsTrajectoryGenerator(object):
  """A trajectory generator that return signals for alternating legs."""

  def __init__(
      self,
      init_abduction=laikago_pose_utils.LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
      init_hip=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
      init_knee=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE,
      amplitude=MOTION_AMPLITUDE,
      frequency=MOTION_FREQUENCY,
      gait=PACE_GAIT,  # can be TROT_GAIT or PACE_GAIT
  ):
    """Initializes the controller."""
    self._pose = np.array(
        attr.astuple(
            laikago_pose_utils.LaikagoPose(
                abduction_angle_0=init_abduction,
                hip_angle_0=init_hip,
                knee_angle_0=init_knee,
                abduction_angle_1=init_abduction,
                hip_angle_1=init_hip,
                knee_angle_1=init_knee,
                abduction_angle_2=init_abduction,
                hip_angle_2=init_hip,
                knee_angle_2=init_knee,
                abduction_angle_3=init_abduction,
                hip_angle_3=init_hip,
                knee_angle_3=init_knee)))
    action_high = np.array([ACTION_BOUND] * NUM_MOTORS_LAIKAGO)
    self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self.amplitude = amplitude
    self.period = 1.0 / frequency
    self.gait = gait

  def _alternating_legs_trajectory(self, t):
    """The reference trajectory of each joint when alternating legs.

    Args:
      t: The time since the latest robot reset.

    Returns:
      An array of 12 desired motor angles.
    """
    phase_in_period = (t % self.period) / self.period
    is_first_half_gait = phase_in_period < 0.5
    if self.gait == TROT_GAIT and is_first_half_gait:
      phases = [0, 1, 1, 0]  # 0 means stance and 1 means retraction.
    elif self.gait == TROT_GAIT and not is_first_half_gait:
      phases = [1, 0, 0, 1]
    elif self.gait == PACE_GAIT and is_first_half_gait:
      phases = [0, 1, 0, 1]
    elif self.gait == PACE_GAIT and not is_first_half_gait:
      phases = [1, 0, 1, 0]
    else:
      raise ValueError("{} gait is not supported in alternating legs.".format(
          self.gait))

    phase_step_center = 0.25 if is_first_half_gait else 0.75
    std = STD_FOR_GAUSSIAN_TRAJECTORY
    # Uses Gaussian instead of sine for gentle foot touch down.
    # The following joint angles are added to self._pose.
    retract_hip_angle = self.amplitude * math.exp(
        -(phase_in_period - phase_step_center) *
        (phase_in_period - phase_step_center) / (std * std))
    retract_knee_angle = -2.0 * retract_hip_angle
    retract_abduction_angle = 0.0
    stance_hip_angle = 0.0
    stance_knee_angle = 0.0
    stance_abduction_angle = 0.0
    angles = []
    for is_retract in phases:
      if is_retract:
        angles.extend([retract_abduction_angle, retract_hip_angle,
                       retract_knee_angle])
      else:
        angles.extend([stance_abduction_angle, stance_hip_angle,
                       stance_knee_angle])
    return np.array(angles)

  def reset(self):
    pass

  def get_action(self, current_time, input_action):
    """Computes the trajectory according to input time and action.

    Args:
      current_time: The time in gym env since reset.
      input_action: A numpy array. The input leg pose from a NN controller.

    Returns:
      A numpy array. The desired motor angles.
    """

    return self._pose + self._alternating_legs_trajectory(
        current_time) + input_action

  def get_observation(self, input_observation):
    """Get the trajectory generator's observation."""

    return input_observation



import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import time
import pybullet as p
from . import kuka
import random
import pybullet_data
from pkg_resources import parse_version

maxSteps = 1000

RENDER_HEIGHT = 720
RENDER_WIDTH = 960


class KukaCamGymEnv(gym.Env):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self,
               urdfRoot=pybullet_data.getDataPath(),
               actionRepeat=1,
               isEnableSelfCollision=True,
               renders=False,
               isDiscrete=False):
    self._timeStep = 1. / 240.
    self._urdfRoot = urdfRoot
    self._actionRepeat = actionRepeat
    self._isEnableSelfCollision = isEnableSelfCollision
    self._observation = []
    self._envStepCounter = 0
    self._renders = renders
    self._width = 341
    self._height = 256
    self._isDiscrete = isDiscrete
    self.terminated = 0
    self._p = p
    if self._renders:
      cid = p.connect(p.SHARED_MEMORY)
      if (cid < 0):
        p.connect(p.GUI)
      p.resetDebugVisualizerCamera(1.3, 180, -41, [0.52, -0.2, -0.33])
    else:
      p.connect(p.DIRECT)
    #timinglog = p.startStateLogging(p.STATE_LOGGING_PROFILE_TIMINGS, "kukaTimings.json")
    self.seed()
    self.reset()
    observationDim = len(self.getExtendedObservation())
    #print("observationDim")
    #print(observationDim)

    observation_high = np.array([np.finfo(np.float32).max] * observationDim)
    if (self._isDiscrete):
      self.action_space = spaces.Discrete(7)
    else:
      action_dim = 3
      self._action_bound = 1
      action_high = np.array([self._action_bound] * action_dim)
      self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self.observation_space = spaces.Box(low=0,
                                        high=255,
                                        shape=(self._height, self._width, 4),
                                        dtype=np.uint8)
    self.viewer = None

  def reset(self):
    self.terminated = 0
    p.resetSimulation()
    p.setPhysicsEngineParameter(numSolverIterations=150)
    p.setTimeStep(self._timeStep)
    p.loadURDF(os.path.join(self._urdfRoot, "plane.urdf"), [0, 0, -1])

    p.loadURDF(os.path.join(self._urdfRoot, "table/table.urdf"), 0.5000000, 0.00000, -.820000,
               0.000000, 0.000000, 0.0, 1.0)

    xpos = 0.5 + 0.2 * random.random()
    ypos = 0 + 0.25 * random.random()
    ang = 3.1415925438 * random.random()
    orn = p.getQuaternionFromEuler([0, 0, ang])
    self.blockUid = p.loadURDF(os.path.join(self._urdfRoot, "block.urdf"), xpos, ypos, -0.1,
                               orn[0], orn[1], orn[2], orn[3])

    p.setGravity(0, 0, -10)
    self._kuka = kuka.Kuka(urdfRootPath=self._urdfRoot, timeStep=self._timeStep)
    self._envStepCounter = 0
    p.stepSimulation()
    self._observation = self.getExtendedObservation()
    return np.array(self._observation)

  def __del__(self):
    p.disconnect()

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def getExtendedObservation(self):

    #camEyePos = [0.03,0.236,0.54]
    #distance = 1.06
    #pitch=-56
    #yaw = 258
    #roll=0
    #upAxisIndex = 2
    #camInfo = p.getDebugVisualizerCamera()
    #print("width,height")
    #print(camInfo[0])
    #print(camInfo[1])
    #print("viewMatrix")
    #print(camInfo[2])
    #print("projectionMatrix")
    #print(camInfo[3])
    #viewMat = camInfo[2]
    #viewMat = p.computeViewMatrixFromYawPitchRoll(camEyePos,distance,yaw, pitch,roll,upAxisIndex)
    viewMat = [
        -0.5120397806167603, 0.7171027660369873, -0.47284144163131714, 0.0, -0.8589617609977722,
        -0.42747554183006287, 0.28186774253845215, 0.0, 0.0, 0.5504802465438843,
        0.8348482847213745, 0.0, 0.1925382763147354, -0.24935829639434814, -0.4401884973049164, 1.0
    ]
    #projMatrix = camInfo[3]#[0.7499999403953552, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1.0000200271606445, -1.0, 0.0, 0.0, -0.02000020071864128, 0.0]
    projMatrix = [
        0.75, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1.0000200271606445, -1.0, 0.0, 0.0,
        -0.02000020071864128, 0.0
    ]

    img_arr = p.getCameraImage(width=self._width,
                               height=self._height,
                               viewMatrix=viewMat,
                               projectionMatrix=projMatrix)
    rgb = img_arr[2]
    np_img_arr = np.reshape(rgb, (self._height, self._width, 4))
    self._observation = np_img_arr
    return self._observation

  def step(self, action):
    if (self._isDiscrete):
      dv = 0.01
      dx = [0, -dv, dv, 0, 0, 0, 0][action]
      dy = [0, 0, 0, -dv, dv, 0, 0][action]
      da = [0, 0, 0, 0, 0, -0.1, 0.1][action]
      f = 0.3
      realAction = [dx, dy, -0.002, da, f]
    else:
      dv = 0.01
      dx = action[0] * dv
      dy = action[1] * dv
      da = action[2] * 0.1
      f = 0.3
      realAction = [dx, dy, -0.002, da, f]

    return self.step2(realAction)

  def step2(self, action):
    for i in range(self._actionRepeat):
      self._kuka.applyAction(action)
      p.stepSimulation()
      if self._termination():
        break
      #self._observation = self.getExtendedObservation()
      self._envStepCounter += 1

    self._observation = self.getExtendedObservation()
    if self._renders:
      time.sleep(self._timeStep)

    #print("self._envStepCounter")
    #print(self._envStepCounter)

    done = self._termination()
    reward = self._reward()
    #print("len=%r" % len(self._observation))

    return np.array(self._observation), reward, done, {}

  def render(self, mode='human', close=False):
    if mode != "rgb_array":
      return np.array([])
    return self.getExtendedObservation()

  def _termination(self):
    #print (self._kuka.endEffectorPos[2])
    state = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaEndEffectorIndex)
    actualEndEffectorPos = state[0]

    #print("self._envStepCounter")
    #print(self._envStepCounter)
    if (self.terminated or self._envStepCounter > maxSteps):
      self._observation = self.getExtendedObservation()
      return True
    maxDist = 0.005
    closestPoints = p.getClosestPoints(self._kuka.trayUid, self._kuka.kukaUid, maxDist)

    if (len(closestPoints)):  #(actualEndEffectorPos[2] <= -0.43):
      self.terminated = 1

      #print("closing gripper, attempting grasp")
      #start grasp and terminate
      fingerAngle = 0.3
      for i in range(100):
        graspAction = [0, 0, 0.0001, 0, fingerAngle]
        self._kuka.applyAction(graspAction)
        p.stepSimulation()
        fingerAngle = fingerAngle - (0.3 / 100.)
        if (fingerAngle < 0):
          fingerAngle = 0

      for i in range(1000):
        graspAction = [0, 0, 0.001, 0, fingerAngle]
        self._kuka.applyAction(graspAction)
        p.stepSimulation()
        blockPos, blockOrn = p.getBasePositionAndOrientation(self.blockUid)
        if (blockPos[2] > 0.23):
          #print("BLOCKPOS!")
          #print(blockPos[2])
          break
        state = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaEndEffectorIndex)
        actualEndEffectorPos = state[0]
        if (actualEndEffectorPos[2] > 0.5):
          break

      self._observation = self.getExtendedObservation()
      return True
    return False

  def _reward(self):

    #rewards is height of target object
    blockPos, blockOrn = p.getBasePositionAndOrientation(self.blockUid)
    closestPoints = p.getClosestPoints(self.blockUid, self._kuka.kukaUid, 1000, -1,
                                       self._kuka.kukaEndEffectorIndex)

    reward = -1000
    numPt = len(closestPoints)
    #print(numPt)
    if (numPt > 0):
      #print("reward:")
      reward = -closestPoints[0][8] * 10
    if (blockPos[2] > 0.2):
      #print("grasped a block!!!")
      #print("self._envStepCounter")
      #print(self._envStepCounter)
      reward = reward + 1000

    #print("reward")
    #print(reward)
    return reward

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



"""Change the action from uv of the depth map to xyz in world frame."""

import copy
import math
from typing import Sequence
import gin
import gym
import numpy as np

from pybullet_envs.minitaur.robots import laikago_kinematic_constants
from pybullet_envs.minitaur.robots import quadruped_base
from pybullet_envs.minitaur.robots import wheeled_robot_base_sim

UNIT_QUATERNION = (0, 0, 0, 1)
# A small gap between the target foot position and the floor.
FLOOR_HEIGHT_EPSILON = 0.02
FOOT_HEIGHT_CLEARNCE = 0.075
BASE_MOVEMENT_SPEED = 0.3
INIT_ABDUCTION_ANGLE = laikago_kinematic_constants.INIT_ABDUCTION_ANGLE
INIT_HIP_ANGLE = laikago_kinematic_constants.INIT_HIP_ANGLE
INIT_KNEE_ANGLE = laikago_kinematic_constants.INIT_KNEE_ANGLE
NUM_LEGS = laikago_kinematic_constants.NUM_LEGS
COM_VIZ_BOX_SIZE = [0.025, 0.025, 0.005]
_DEFAULT_JOINT_POSE = (INIT_ABDUCTION_ANGLE, INIT_HIP_ANGLE,
                       INIT_KNEE_ANGLE) * NUM_LEGS
# Weights for computing the target COM from the supporting feet locations.
# The target COM for the front feet are biased forward for better robot
# stability and seeing farther in distance, which enables larger steps.
_SUPPORT_WEIGHT_MAP = [[0.0, 0.4, 0.3, 0.3], [0.4, 0.0, 0.3, 0.3],
                       [1.0 / 3, 1.0 / 3, 0.0, 1.0 / 3],
                       [1.0 / 3, 1.0 / 3, 1.0 / 3, 0.0]]
_TASK_SENSOR_NAME = "sensors"


@gin.configurable
class DepthUVToFootPlacementWrapper(object):
  """Changes the action from a point in depth map to a point in the world frame.

  Attributes:
    observation: The current observation of the environment.
    last_action: The last that was used to step the environment.
    env_step_counter: The number of control steps that have been elapesed since
      the environment is reset.
    action_space: The action space of the environment.
  """

  def __init__(self,
               gym_env,
               visualization=True,
               foot_movement=False,
               foot_clearance_height=FOOT_HEIGHT_CLEARNCE,
               base_movement_speed=BASE_MOVEMENT_SPEED,
               foothold_update_frequency=4):
    """Initializes the wrapper.

    Args:
      gym_env: the wrapped gym environment. The robot is controlled
        kinematically when gym_env.robot inherits from WheeledRobotBase and is
        controlled dynamically using a static gait controller when gym_env.robot
        inherits from QudrupedBase.
      visualization: whether to draw a sphere that represents the foothold
        position.
      foot_movement: whether move the toe to the desired foothold position using
        IK for visualization/debugging purpose.
      foot_clearance_height: the maximum height of the swing foot.
      base_movement_speed: the speed of the robot base.
      foothold_update_frequency: the frequency of updating the foothold, which
        is the same as the frequency of the steps. The default value 4 means
        four steps (1 complete cycle of static gait) per second.
    """
    self._gym_env = gym_env
    self._num_control_steps_per_foothold_update = max(
        1, int(1.0 / foothold_update_frequency / self._gym_env.env_time_step))
    self.last_action = None
    self._visualization = visualization
    self._foot_movement = foot_movement
    self._foot_clearance_height = foot_clearance_height
    self._base_movement_speed = base_movement_speed
    self._step_counter = 0
    self.observation_space.spaces["toe_position"] = gym.spaces.Box(
        np.array([-1.0] * 3), np.array([1.0] * 3))
    self.action_space = gym.spaces.Box(
        np.array([-1.0] * 2), np.array([1.0] * 2))
    self.task.reset(self)
    if hasattr(self.task, _TASK_SENSOR_NAME):
      self.observation_space.spaces[
          self.task.get_name()] = self.task.observation_space

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def reset(self, **kwargs):
    """Reset the environment."""
    obs = self._gym_env.reset(**kwargs)
    self.task.reset(self)
    self._step_counter = 0
    current_end_effector_pos = np.array(
        self._gym_env.robot.foot_positions()[self.task.swing_foot_id])
    self.last_action = [0, 0]
    if self._visualization:
      self._create_foot_placement_visualization()
      self._create_com_visualization()
    self._initial_local_toe_positions = np.array(
        self._gym_env.robot.foot_positions(position_in_world_frame=False))

    # Move COM to prepare for the first swing step.
    if isinstance(self._gym_env.robot, quadruped_base.QuadrupedBase):
      obs, _ = self._move_com_dynamic(
          self.task.swing_foot_id,
          self._num_control_steps_per_foothold_update // 3 * 2)

    # TODO(b/157614175): Adds a toe_position_sensor.
    obs["toe_position"] = current_end_effector_pos
    obs["vision"] = self.task.get_depth_image_for_foot()
    obs["LastAction"] = self.last_action

    self._observation = obs
    return self._observation

  def _move_kinematic(self):
    """Move robot kinematically.

    Returns:
      The tuple containing the observation and env info.
    """
    if self._foot_movement:
      toe_positions = np.array(
          self._gym_env.robot.foot_positions(position_in_world_frame=True))
      toe_positions[:, 2] = FLOOR_HEIGHT_EPSILON
      destination_foothold_xyz_global = self.task.get_foothold_location(
          self.last_action, use_world_frame=True)
      destination_foothold_xyz_global[2] = FLOOR_HEIGHT_EPSILON
    joint_pose = _DEFAULT_JOINT_POSE
    for i in range(self._num_control_steps_per_foothold_update):
      if self._foot_movement:
        alpha = i / (self._num_control_steps_per_foothold_update - 1)
        toe_positions_over_time = copy.deepcopy(toe_positions)
        toe_positions_over_time[self.task.swing_foot_id] = (
            self._construct_foot_trajectories(
                alpha, toe_positions[self.task.swing_foot_id],
                destination_foothold_xyz_global, self._foot_clearance_height))
        joint_pose = np.array(
            self.robot.motor_angles_from_foot_positions(
                toe_positions_over_time, position_in_world_frame=True)[1])
      action = {
          wheeled_robot_base_sim.BASE_ACTION_NAME:
              (self._base_movement_speed, 0),
          wheeled_robot_base_sim.BODY_ACTION_NAME:
              joint_pose,
      }
      obs, _, _, info = self._gym_env.step(action)
    return obs, info

  def _move_com_dynamic(self, swing_foot_id, num_control_steps):
    """Move robot COM to the weightd center of the support polygon through dynamics.

    Args:
      swing_foot_id: Index of the swing foot.
      num_control_steps: Total number of control steps for moving the COM.

    Returns:
      The tuple containing the observation and env info.
    """

    toe_positions = np.array(
        self._gym_env.robot.foot_positions(position_in_world_frame=False))
    support_polygon_center = np.array([0.0, 0.0, 0.0])
    for i in range(NUM_LEGS):
      if i != swing_foot_id:
        support_polygon_center += toe_positions[i] * _SUPPORT_WEIGHT_MAP[
            swing_foot_id][i]
    support_polygon_center[2] = 0.0

    for i in range(num_control_steps):
      alpha = i / (num_control_steps - 1)
      # Create a flat phase towards the end to stabilize the com movement.
      alpha = np.clip(alpha * 1.1, 0, 1)
      toe_positions_over_time = copy.deepcopy(
          toe_positions) - alpha * support_polygon_center
      # Use initial toe height to maintain the overal base height.
      for j in range(len(toe_positions_over_time)):
        toe_positions_over_time[j][2] = self._initial_local_toe_positions[j][2]
      joint_pose = np.array(
          self.robot.motor_angles_from_foot_positions(
              toe_positions_over_time, position_in_world_frame=False)[1])
      obs, _, _, info = self._gym_env.step(joint_pose)
      self._update_com_visualization()
    return obs, info

  def _swing_leg_dynamic(self, swing_foot_id, destination_foothold_xyz_local,
                         num_control_steps):
    """Move swing leg to the target foothold through IK and dynamics.

    Args:
      swing_foot_id: Index of the swing foot.
      destination_foothold_xyz_local: Target foothold position.
      num_control_steps: Total number of control steps for swinging the leg.

    Returns:
      The tuple containing the observation and env info.
    """

    local_toe_positions = np.array(
        self._gym_env.robot.foot_positions(position_in_world_frame=False))

    # Move swing leg
    for i in range(num_control_steps):
      alpha = i / (num_control_steps - 1)
      toe_positions_over_time = copy.deepcopy(local_toe_positions)
      toe_positions_over_time[swing_foot_id] = (
          self._construct_foot_trajectories(alpha,
                                            local_toe_positions[swing_foot_id],
                                            destination_foothold_xyz_local,
                                            self._foot_clearance_height))
      joint_pose = np.array(
          self.robot.motor_angles_from_foot_positions(
              toe_positions_over_time, position_in_world_frame=False)[1])

      obs, _, _, info = self._gym_env.step(joint_pose)
    return obs, info

  def _move_dynamic(self):
    """Move robot dynamically.

    Returns:
      The tuple containing the observation and env info.
    """
    destination_foothold_xyz_local = self.task.get_foothold_location(
        self.last_action, use_world_frame=False)
    # Lift the target foothold slightly to account for thickness of the feet.
    destination_foothold_xyz_local[2] += FLOOR_HEIGHT_EPSILON

    # Swing leg in the first 1/3 of the duration.
    self._swing_leg_dynamic(self.task.swing_foot_id,
                            destination_foothold_xyz_local,
                            self._num_control_steps_per_foothold_update // 3)

    # Move COM in the rest 2/3 of the duration.
    obs, info = self._move_com_dynamic(
        self.task.next_swing_foot_id,
        self._num_control_steps_per_foothold_update // 3 * 2)

    return obs, info

  def step(self, action: Sequence[float]):
    """Steps the wrapped environment.

    Args:
      action: 2 dimensional numpy array between [-1.0, 1.0]. They represents the
        depth image pixel index. We assume that only one foot is swinging in
        this wrapper, and this is the foothold location for that swinging leg.
        The order of the swinging leg and the index of the current swinging leg
        is defined in stepstone_visiontask.py.

    Returns:
      The tuple containing the observation, the reward, and the episode
        end indicator.
    """
    self.last_action = action
    reward = self.task(self)
    done = self.task.done(self)
    if self._visualization:
      self._update_foothold_visualization(action)

    if isinstance(self._gym_env.robot, quadruped_base.QuadrupedBase):
      obs, info = self._move_dynamic()
    else:
      obs, info = self._move_kinematic()

    self._step_counter += 1
    current_end_effector_pos = np.array(
        self._gym_env.robot.foot_positions()[self.task.swing_foot_id])
    obs["toe_position"] = current_end_effector_pos
    obs["vision"] = self.task.get_depth_image_for_foot()
    self._observation = obs
    return obs, reward, done, info

  def _construct_foot_trajectories(self, alpha, swing_foot_start_position,
                                   swing_foot_destination,
                                   foot_clearance_height):
    """Construct the target foot position for the swing foot.

    Args:
      alpha: a float in [0.0, 1.0] indicating the phase of the swing foot.
      swing_foot_start_position: foot position at the beginning of the swing
        motion.
      swing_foot_destination: target foot position at the end of the swing
        motion.
      foot_clearance_height: the maximum height of the swing foot.

    Returns:
      The interpolated swing foot position.
    """
    new_swing_foot_position = swing_foot_start_position + alpha * (
        swing_foot_destination - swing_foot_start_position)
    new_swing_foot_position[2] += (
        foot_clearance_height * math.sin((alpha) * math.pi))
    return new_swing_foot_position

  def _create_foot_placement_visualization(self):
    """Creates a visualization sphere that represents the foothold position."""
    visual_id = self._gym_env.pybullet_client.createVisualShape(
        self._gym_env.pybullet_client.GEOM_SPHERE,
        radius=0.02,
        rgbaColor=[0.7, 0.7, 0.7, 1])
    self._foothold_visual_body = self._gym_env.pybullet_client.createMultiBody(
        baseMass=0, baseVisualShapeIndex=visual_id, basePosition=[0, 0, 0])

  def _update_foothold_visualization(self, action):
    """Moves the visualization sphere that represents the next foothold."""
    foothold_xyz = self.task.get_foothold_location(
        action, use_world_frame=True)
    self._gym_env.pybullet_client.resetBasePositionAndOrientation(
        self._foothold_visual_body, foothold_xyz, UNIT_QUATERNION)

  def _create_com_visualization(self):
    """Creates visualization boxes for COM and support polygon center."""
    visual_id = self._gym_env.pybullet_client.createVisualShape(
        self._gym_env.pybullet_client.GEOM_BOX,
        halfExtents=COM_VIZ_BOX_SIZE,
        rgbaColor=[0.7, 0.5, 0.5, 1])
    self._support_polygon_center_visual_body = self._gym_env.pybullet_client.createMultiBody(
        baseMass=0, baseVisualShapeIndex=visual_id, basePosition=[0, 0, 0])

    visual_id = self._gym_env.pybullet_client.createVisualShape(
        self._gym_env.pybullet_client.GEOM_BOX,
        halfExtents=COM_VIZ_BOX_SIZE,
        rgbaColor=[0.5, 0.75, 0.5, 1])
    self._projected_com_visual_body = self._gym_env.pybullet_client.createMultiBody(
        baseMass=0, baseVisualShapeIndex=visual_id, basePosition=[0, 0, 0])

  def _update_com_visualization(self):
    """Moves the visualization for the COM and support polygon center."""
    toe_positions = np.array(
        self._gym_env.robot.foot_positions(position_in_world_frame=True))
    support_polygon_center_global = np.array([0.0, 0.0, 0.0])
    for i in range(NUM_LEGS):
      if i != self.task.next_swing_foot_id:
        support_polygon_center_global += toe_positions[i] * _SUPPORT_WEIGHT_MAP[
            self.task.next_swing_foot_id][i]
    support_polygon_center_global[2] = 0.0
    self._gym_env.pybullet_client.resetBasePositionAndOrientation(
        self._support_polygon_center_visual_body, support_polygon_center_global,
        UNIT_QUATERNION)

    com = np.copy(self._gym_env.robot.base_position)
    com[2] = 0.0
    self._gym_env.pybullet_client.resetBasePositionAndOrientation(
        self._projected_com_visual_body, com, UNIT_QUATERNION)

  @property
  def observation(self):
    return self._observation

  @property
  def env_step_counter(self):
    return self._step_counter



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute a streaming estimation of the mean of submitted tensors."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf


class StreamingMean(object):
  """Compute a streaming estimation of the mean of submitted tensors."""

  def __init__(self, shape, dtype):
    """Specify the shape and dtype of the mean to be estimated.

    Note that a float mean to zero submitted elements is NaN, while computing
    the integer mean of zero elements raises a division by zero error.

    Args:
      shape: Shape of the mean to compute.
      dtype: Data type of the mean to compute.
    """
    self._dtype = dtype
    self._sum = tf.Variable(lambda: tf.zeros(shape, dtype), False)
    self._count = tf.Variable(lambda: 0, trainable=False)

  @property
  def value(self):
    """The current value of the mean."""
    return self._sum / tf.cast(self._count, self._dtype)

  @property
  def count(self):
    """The number of submitted samples."""
    return self._count

  def submit(self, value):
    """Submit a single or batch tensor to refine the streaming mean."""
    # Add a batch dimension if necessary.
    if value.shape.ndims == self._sum.shape.ndims:
      value = value[None, ...]
    return tf.group(self._sum.assign_add(tf.reduce_sum(value, 0)),
                    self._count.assign_add(tf.shape(value)[0]))

  def clear(self):
    """Return the mean estimate and reset the streaming statistics."""
    value = self._sum / tf.cast(self._count, self._dtype)
    with tf.control_dependencies([value]):
      reset_value = self._sum.assign(tf.zeros_like(self._sum))
      reset_count = self._count.assign(0)
    with tf.control_dependencies([reset_value, reset_count]):
      return tf.identity(value)



"""Definitions of safety layer data types."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import attr


@attr.s
class Bound(object):
  """Struct for inclusive lower and upper bounds."""
  lower = attr.ib(type=float, default=0)
  upper = attr.ib(type=float, default=0)

  @upper.validator  # pytype: disable=attribute-error
  def _upper_greator_equal_to_lower(self, attribute, value):
    del attribute
    assert value >= self.lower, (
        "upper bound {} is less than lower bound {}".format(value, self.lower))


@attr.s
class SafetyConfig(object):
  """Struct to configure the safety module."""
  motor_position_bound = attr.ib(type=list)
  motor_position_gain_bound = attr.ib(type=list)
  motor_velocity_bound = attr.ib(type=list)
  motor_velocity_gain_bound = attr.ib(type=list)
  motor_torque_bound = attr.ib(type=list)
  timestamp_delta_bound = attr.ib(type=Bound)
  motor_average_abs_velocity_bound = attr.ib(type=list)
  motor_average_abs_power_bound = attr.ib(type=list)
  state_action_timestamp_delta_bound = attr.ib(type=float)
  motor_delta_position_bound = attr.ib(type=list)
  motor_average_abs_delta_position_bound = attr.ib(type=list)


@attr.s
class MotorState(object):
  """A generic type for motor state.

  Motor states are what we can potentially read from the motor encoder or
  firmware APIs.

  """
  timestamp = attr.ib(type=float, default=None)
  position = attr.ib(type=float, default=None)
  position_gain = attr.ib(type=float, default=None)
  velocity = attr.ib(type=float, default=None)
  velocity_gain = attr.ib(type=float, default=None)
  torque = attr.ib(type=float, default=None)


@attr.s
class MotorAction(object):
  """A generic type for motor action.

   Motor actions are the potential command structure we can send to the motor.
   While similar to MotorState, they are logically very different entities.
  """
  timestamp = attr.ib(type=float, default=None)
  position = attr.ib(type=float, default=None)
  position_gain = attr.ib(type=float, default=None)
  velocity = attr.ib(type=float, default=None)
  velocity_gain = attr.ib(type=float, default=None)
  torque = attr.ib(type=float, default=None)



# Lint as: python3
"""The system to log and manage all metrics."""

import logging
from typing import Any, Callable, Dict, Sequence, Text, Union
import numpy as np

from pybullet_envs.minitaur.envs_v2.evaluation import metric as metric_lib


def _merge_dict_throw_on_duplicates(
    dict_1: Dict[Text, Any],
    dict_2: Dict[Text, Any],
):
  """Merge the contents of dict_2 to dict_1.

  Args:
    dict_1: The dictionary to merge into.
    dict_2: The dictionary to merge from.

  Raises:
    KeyError: if duplicated keys are found.
  """
  for key, val in dict_2.items():
    if key in dict_1:
      raise KeyError(f"Duplicate key: {key} found in both "
                     f"dictionaries: {dict_1} {dict_2}.")
    dict_1[key] = val


def common_stats(x: Sequence[Union[float, Sequence[float]]],
                 flatten_array=False):
  out = x
  if flatten_array:
    # Deals with array of arrays.
    out = np.concatenate(x).flatten()
  return {
      "mean": np.mean(out),
      "max": np.max(out),
      "min": np.min(out),
      "std": np.std(out),
  }


class MetricLogger(object):
  """The central system to manage all metrics."""

  def __init__(self, ignore_duplicate_metrics: bool = True):
    """Initializes the system.

    Args:
      ignore_duplicate_metrics: Don't throw error when users want to recreate
        the same metric.
    """
    self._metric_reporters = {}
    self._ignore_duplicate_metrics = ignore_duplicate_metrics

  def reset_episode(self):
    """Resets all metric reporters's internal buffer.

    Will be called by the gym env during reset.
    """
    for metric in self._metric_reporters.values():
      metric.reset_episode()

  def create_metric(
      self,
      name: Text,
      scope: metric_lib.MetricScope,
      single_ep_aggregator: Callable[[Sequence[Any]], Any],
      multi_ep_aggregator: Callable[[Sequence[Any]], Dict[Text, Any]],
  ) -> metric_lib.MetricCore:
    """Creates a new metric.

    Args:
      name: The name of the metric, for example "motor_torques",
        "distance_to_wall", etc. The full name of the metric will have scope
        name in the prefix, i.e. "scope/name".
      scope: The scope of this metric. Most metric should be for DEBUG purpose.
        The scope name will be added to the final name of metric in this way:
          "scope/name", which is standarded format for Tensorboard to group
          named variables.
      single_ep_aggregator: The function to process all aggregated metric
        values. The derived MetricReporter (see below) will implements
        reset_episode() which clears the episode data, and will be called during
        env.reset().
      multi_ep_aggregator: The function to process multi-episode metric values.
        We assume the inputs to the function is a list of per episode metric
        values, i.e. each element of the list is the output from the
        single_ep_aggregator.

    Returns:
      A MetricCore which can be used to report values of interest.
    """
    if name in self._metric_reporters:
      if self._ignore_duplicate_metrics:
        logging.warning("Trying to create an existing metric: %s", name)
        name = self._get_valid_duplicate_name(name)
        logging.warning("Changed duplicate metric to new name: %s", name)
      else:
        raise ValueError(f"Duplicated metrics found: {name}")

    self._metric_reporters[name] = metric_lib.MetricReporter(
        name, scope, single_ep_aggregator, multi_ep_aggregator)
    return self._metric_reporters[name]

  def create_scalar_metric(
      self,
      name: Text,
      scope: metric_lib.MetricScope,
      single_ep_aggregator: Callable[[Sequence[Any]], Any],
  ) -> metric_lib.MetricCore:
    """Shortcut to create a metric for scalar variables."""
    return self.create_metric(
        name,
        scope,
        single_ep_aggregator,
        multi_ep_aggregator=common_stats,
    )

  def _get_valid_duplicate_name(self, original_name: Text) -> Text:
    counter = 1
    test_name = "{}_duplicate_{}".format(original_name, str(counter))
    while test_name in self._metric_reporters:
      counter += 1
      test_name = "{}_duplicate_{}".format(original_name, str(counter))
    return test_name

  def get_episode_metrics(self) -> Dict[Text, Any]:
    """Return all metrics registered in the logger for the current episode."""
    ep_stats = {}
    for metric in self._metric_reporters.values():
      _merge_dict_throw_on_duplicates(ep_stats, metric.get_episode_metric())
    return ep_stats

  def get_multi_episode_metrics(
      self, episodic_metrics: Dict[Text, Sequence[Any]]) -> Dict[Text, Any]:
    """Processes the aggregated metrics over many episodes.

    Will not be affected by reset_episode, since we take multi-episode data as
    inputs.

    Args:
      episodic_metrics: The per episode metrics. For each key in the inputs, we
        expect at least one metric reporter can process the corresponding value,
        which contains a list of episodic metrics.

    Returns:
      The processed multi-episode metrics.
    """
    stats = {}
    for metric in self._metric_reporters.values():
      _merge_dict_throw_on_duplicates(
          stats, metric.get_multi_ep_metric(episodic_metrics))
    return stats



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for using reinforcement learning algorithms."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
import re

import ruamel.yaml as yaml
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from . import tools


def define_simulation_graph(batch_env, algo_cls, config):
  """Define the algortihm and environment interaction.

  Args:
    batch_env: In-graph environments object.
    algo_cls: Constructor of a batch algorithm.
    config: Configuration object for the algorithm.

  Returns:
    Object providing graph elements via attributes.
  """
  # pylint: disable=unused-variable
  step = tf.Variable(0, False, dtype=tf.int32, name='global_step')
  is_training = tf.placeholder(tf.bool, name='is_training')
  should_log = tf.placeholder(tf.bool, name='should_log')
  do_report = tf.placeholder(tf.bool, name='do_report')
  force_reset = tf.placeholder(tf.bool, name='force_reset')
  algo = algo_cls(batch_env, step, is_training, should_log, config)
  done, score, summary = tools.simulate(batch_env, algo, should_log, force_reset)
  message = 'Graph contains {} trainable variables.'
  tf.logging.info(message.format(tools.count_weights()))
  # pylint: enable=unused-variable
  return tools.AttrDict(locals())


def define_batch_env(constructor, num_agents, env_processes):
  """Create environments and apply all desired wrappers.

  Args:
    constructor: Constructor of an OpenAI gym environment.
    num_agents: Number of environments to combine in the batch.
    env_processes: Whether to step environment in external processes.

  Returns:
    In-graph environments object.
  """
  with tf.variable_scope('environments'):
    if env_processes:
      envs = [tools.wrappers.ExternalProcess(constructor) for _ in range(num_agents)]
    else:
      envs = [constructor() for _ in range(num_agents)]
    batch_env = tools.BatchEnv(envs, blocking=not env_processes)
    batch_env = tools.InGraphBatchEnv(batch_env)
  return batch_env


def define_saver(exclude=None):
  """Create a saver for the variables we want to checkpoint.

  Args:
    exclude: List of regexes to match variable names to exclude.

  Returns:
    Saver object.
  """
  variables = []
  exclude = exclude or []
  exclude = [re.compile(regex) for regex in exclude]
  for variable in tf.global_variables():
    if any(regex.match(variable.name) for regex in exclude):
      continue
    variables.append(variable)
  saver = tf.train.Saver(variables, keep_checkpoint_every_n_hours=5)
  return saver


def initialize_variables(sess, saver, logdir, checkpoint=None, resume=None):
  """Initialize or restore variables from a checkpoint if available.

  Args:
    sess: Session to initialize variables in.
    saver: Saver to restore variables.
    logdir: Directory to search for checkpoints.
    checkpoint: Specify what checkpoint name to use; defaults to most recent.
    resume: Whether to expect recovering a checkpoint or starting a new run.

  Raises:
    ValueError: If resume expected but no log directory specified.
    RuntimeError: If no resume expected but a checkpoint was found.
  """
  sess.run(tf.group(tf.local_variables_initializer(), tf.global_variables_initializer()))
  if resume and not (logdir or checkpoint):
    raise ValueError('Need to specify logdir to resume a checkpoint.')
  if logdir:
    state = tf.train.get_checkpoint_state(logdir)
    if checkpoint:
      checkpoint = os.path.join(logdir, checkpoint)
    if not checkpoint and state and state.model_checkpoint_path:
      checkpoint = state.model_checkpoint_path
    if checkpoint and resume is False:
      message = 'Found unexpected checkpoint when starting a new run.'
      raise RuntimeError(message)
    if checkpoint:
      saver.restore(sess, checkpoint)


def save_config(config, logdir=None):
  """Save a new configuration by name.

  If a logging directory is specified, is will be created and the configuration
  will be stored there. Otherwise, a log message will be printed.

  Args:
    config: Configuration object.
    logdir: Location for writing summaries and checkpoints if specified.

  Returns:
    Configuration object.
  """
  if logdir:
    with config.unlocked:
      config.logdir = logdir
    message = 'Start a new run and write summaries and checkpoints to {}.'
    tf.logging.info(message.format(config.logdir))
    tf.gfile.MakeDirs(config.logdir)
    config_path = os.path.join(config.logdir, 'config.yaml')
    with tf.gfile.GFile(config_path, 'w') as file_:
      yaml.dump(config, file_, default_flow_style=False)
  else:
    message = ('Start a new run without storing summaries and checkpoints since no '
               'logging directory was specified.')
    tf.logging.info(message)
  return config


def load_config(logdir):
  """Load a configuration from the log directory.

  Args:
    logdir: The logging directory containing the configuration file.

  Raises:
    IOError: The logging directory does not contain a configuration file.

  Returns:
    Configuration object.
  """
  config_path = logdir and os.path.join(logdir, 'config.yaml')
  if not config_path or not tf.gfile.Exists(config_path):
    message = ('Cannot resume an existing run since the logging directory does not '
               'contain a configuration file.')
    raise IOError(message)
  with tf.gfile.FastGFile(config_path, 'r') as file_:
    config = yaml.load(file_, Loader=yaml.Loader)
  message = 'Resume run and write summaries and checkpoints to {}.'
  tf.logging.info(message.format(config.logdir))
  return config


def set_up_logging():
  """Configure the TensorFlow logger."""
  tf.logging.set_verbosity(tf.logging.INFO)
  logging.getLogger('tensorflow').propagate = False



"""Extends the environment by adding observation and action history.

The implementation is a bit dirty import of the implementation in
the experimental branch.

"""

from gym import spaces
import numpy as np

from pybullet_envs.minitaur.envs.minitaur_reactive_env import MinitaurReactiveEnv


class MinitaurExtendedEnv(MinitaurReactiveEnv):
  """The 'extended' environment for Markovian property.

  This class implements to include prior actions and observations to the
  observation vector, thus making the environment "more" Markovian. This is
  especially useful for systems with latencies.

  Args:
    history_length: the length of the historic data
    history_include_actions: a flag for including actions as history
    history_include_states: a flag for including states as history
    include_state_difference: a flag for including the first-order differences
      as history
    include_second_state_difference: a flag for including the second-order state
      differences as history.
    include_base_position: a flag for including the base as observation,
    never_terminate: if this is on, the environment unconditionally never
      terminates.
    action_scale: the scale of actions,
  """
  MAX_BUFFER_SIZE = 1001
  ACTION_DIM = 8
  PARENT_OBSERVATION_DIM = 12
  INIT_EXTENSION_POS = 2.0
  INIT_SWING_POS = 0.0

  metadata = {
      "render.modes": ["human", "rgb_array"],
      "video.frames_per_second": 50,
  }

  def __init__(self,
               history_length=1,
               history_include_actions=True,
               history_include_states=False,
               include_state_difference=False,
               include_second_state_difference=False,
               include_base_position=False,
               include_leg_model=False,
               never_terminate=False,
               action_scale=0.5,
               **kwargs):
    self._kwargs = kwargs

    self._history_length = history_length
    self._history_include_actions = history_include_actions
    self._history_include_states = history_include_states
    self._include_state_difference = include_state_difference
    self._include_second_state_difference = include_second_state_difference
    self._include_base_position = include_base_position
    self._include_leg_model = include_leg_model

    self._never_terminate = never_terminate
    self._action_scale = action_scale

    self._past_parent_observations = np.zeros((self.MAX_BUFFER_SIZE + 1,
                                               self.PARENT_OBSERVATION_DIM))
    self._past_motor_angles = np.zeros((self.MAX_BUFFER_SIZE + 1, 8))
    self._past_actions = np.zeros((self.MAX_BUFFER_SIZE, self.ACTION_DIM))
    self._counter = 0

    super(MinitaurExtendedEnv, self).__init__(**kwargs)
    self.action_space = spaces.Box(-1.0, 1.0, self.action_space.shape)
    self.observation_space = spaces.Box(-np.inf, np.inf,
                                        self._get_observation().shape)
    # This is mainly for the TF-Agents compatibility
    self.action_space.flat_dim = len(self.action_space.low)
    self.observation_space.flat_dim = len(self.observation_space.low)

  def _get_observation(self):
    """Maybe concatenate motor velocity and torque into observations."""
    parent_observation = super(MinitaurExtendedEnv, self)._get_observation()
    parent_observation = np.array(parent_observation)
    # Base class might require this.
    self._observation = parent_observation
    self._past_parent_observations[self._counter] = parent_observation
    num_motors = self.minitaur.num_motors
    self._past_motor_angles[self._counter] = parent_observation[-num_motors:]

    history_states = []
    history_actions = []
    for i in range(self._history_length):
      t = max(self._counter - i - 1, 0)

      if self._history_include_states:
        history_states.append(self._past_parent_observations[t])

      if self._history_include_actions:
        history_actions.append(self._past_actions[t])

    t = self._counter
    tm, tmm = max(0, self._counter - 1), max(0, self._counter - 2)

    state_difference, second_state_difference = [], []
    if self._include_state_difference:
      state_difference = [
          self._past_motor_angles[t] - self._past_motor_angles[tm]
      ]
    if self._include_second_state_difference:
      second_state_difference = [
          self._past_motor_angles[t] - 2 * self._past_motor_angles[tm] +
          self._past_motor_angles[tmm]
      ]

    base_position = []
    if self._include_base_position:
      base_position = np.array((self.minitaur.GetBasePosition()))

    leg_model = []
    if self._include_leg_model:
      raw_motor_angles = self.minitaur.GetMotorAngles()
      leg_model = self.convert_to_leg_model(raw_motor_angles)

    observation_list = (
        [parent_observation] + history_states + history_actions +
        state_difference + second_state_difference + [base_position] +
        [leg_model])

    full_observation = np.concatenate(observation_list)
    return full_observation

  def reset(self):
    """Resets the time and history buffer."""
    self._counter = 0
    self._signal(self._counter)  # This sets the current phase
    self._past_parent_observations = np.zeros((self.MAX_BUFFER_SIZE + 1,
                                               self.PARENT_OBSERVATION_DIM))
    self._past_motor_angles = np.zeros((self.MAX_BUFFER_SIZE + 1, 8))
    self._past_actions = np.zeros((self.MAX_BUFFER_SIZE, self.ACTION_DIM))
    self._counter = 0

    return np.array(super(MinitaurExtendedEnv, self).reset())

  def step(self, action):
    """Step function wrapper can be used to add shaping terms to the reward.

    Args:
      action: an array of the given action

    Returns:
      next_obs: the next observation
      reward: the reward for this experience tuple
      done: the terminal flag
      info: an additional information
    """

    action *= self._action_scale

    self._past_actions[self._counter] = action
    self._counter += 1

    next_obs, _, done, info = super(MinitaurExtendedEnv, self).step(action)

    reward = self.reward()
    info.update(base_reward=reward)

    return next_obs, reward, done, info

  def terminate(self):
    """The helper function to terminate the environment."""
    super(MinitaurExtendedEnv, self)._close()

  def _termination(self):
    """Determines whether the env is terminated or not.

    checks whether 1) the front leg is bent too much or 2) the time exceeds
    the manually set weights.

    Returns:
      terminal: the terminal flag whether the env is terminated or not
    """
    if self._never_terminate:
      return False

    leg_model = self.convert_to_leg_model(self.minitaur.GetMotorAngles())
    swing0 = leg_model[0]
    swing1 = leg_model[2]
    maximum_swing_angle = 0.8
    if swing0 > maximum_swing_angle or swing1 > maximum_swing_angle:
      return True

    if self._counter >= 500:
      return True

    return False

  def reward(self):
    """Compute rewards for the given time step.

    It considers two terms: 1) forward velocity reward and 2) action
    acceleration penalty.

    Returns:
      reward: the computed reward.
    """
    current_base_position = self.minitaur.GetBasePosition()
    dt = self.control_time_step
    velocity = (current_base_position[0] - self._last_base_position[0]) / dt
    velocity_reward = np.clip(velocity, -0.5, 0.5)

    action = self._past_actions[self._counter - 1]
    prev_action = self._past_actions[max(self._counter - 2, 0)]
    prev_prev_action = self._past_actions[max(self._counter - 3, 0)]
    acc = action - 2 * prev_action + prev_prev_action
    action_acceleration_penalty = np.mean(np.abs(acc))

    reward = 0.0
    reward += 1.0 * velocity_reward
    reward -= 0.1 * action_acceleration_penalty

    return reward

  @staticmethod
  def convert_to_leg_model(motor_angles):
    """A helper function to convert motor angles to leg model.

    Args:
      motor_angles: raw motor angles:

    Returns:
      leg_angles: the leg pose model represented in swing and extension.
    """
    # TODO(sehoonha): clean up model conversion codes
    num_legs = 4
    # motor_angles = motor_angles / (np.pi / 4.)
    leg_angles = np.zeros(num_legs * 2)
    for i in range(num_legs):
      motor1, motor2 = motor_angles[2 * i:2 * i + 2]
      swing = (-1)**(i // 2) * 0.5 * (motor2 - motor1)
      extension = 0.5 * (motor1 + motor2)

      leg_angles[i] = swing
      leg_angles[i + num_legs] = extension

    return leg_angles

  def __getstate__(self):
    """A helper get state function for pickling."""
    return {"kwargs": self._kwargs}

  def __setstate__(self, state):
    """A helper set state function for pickling."""
    self.__init__(**state["kwargs"])



import pybullet as p
import json
import time
import pybullet_data


useGUI = True
if useGUI:
  p.connect(p.GUI)
else:
  p.connect(p.DIRECT)
p.setAdditionalSearchPath(pybullet_data.getDataPath())

useZUp = False
useYUp = not useZUp
showJointMotorTorques = False

if useYUp:
  p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP, 1)

from pybullet_examples.pdControllerExplicit import PDControllerExplicitMultiDof
from pybullet_examples.pdControllerStable import PDControllerStableMultiDof

explicitPD = PDControllerExplicitMultiDof(p)
stablePD = PDControllerStableMultiDof(p)

p.resetDebugVisualizerCamera(cameraDistance=7.4,
                             cameraYaw=-94,
                             cameraPitch=-14,
                             cameraTargetPosition=[0.24, -0.02, -0.09])

import pybullet_data
p.setTimeOut(10000)
useMotionCapture = False
useMotionCaptureReset = False  #not	useMotionCapture
useExplicitPD = True

p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.setPhysicsEngineParameter(numSolverIterations=30)
#p.setPhysicsEngineParameter(solverResidualThreshold=1e-30)

#explicit	PD control requires	small	timestep
timeStep = 1. / 600.
#timeStep	=	1./240.

p.setPhysicsEngineParameter(fixedTimeStep=timeStep)

path = pybullet_data.getDataPath() + "/data/motions/humanoid3d_backflip.txt"
#path	=	pybullet_data.getDataPath()+"/data/motions/humanoid3d_cartwheel.txt"
#path	=	pybullet_data.getDataPath()+"/data/motions/humanoid3d_walk.txt"

#p.loadURDF("plane.urdf",[0,0,-1.03])
print("path	=	", path)
with open(path, 'r') as f:
  motion_dict = json.load(f)
#print("motion_dict	=	", motion_dict)
print("len motion=", len(motion_dict))
print(motion_dict['Loop'])
numFrames = len(motion_dict['Frames'])
print("#frames = ", numFrames)

frameId = p.addUserDebugParameter("frame", 0, numFrames - 1, 0)

erpId = p.addUserDebugParameter("erp", 0, 1, 0.2)

kpMotorId = p.addUserDebugParameter("kpMotor", 0, 1, .2)
forceMotorId = p.addUserDebugParameter("forceMotor", 0, 2000, 1000)

jointTypes = [
    "JOINT_REVOLUTE", "JOINT_PRISMATIC", "JOINT_SPHERICAL", "JOINT_PLANAR", "JOINT_FIXED"
]

startLocations = [[0, 0, 2], [0, 0, 0], [0, 0, -2], [0, 0, -4], [0, 0, 4]]

p.addUserDebugText("Stable PD",
                   [startLocations[0][0], startLocations[0][1] + 1, startLocations[0][2]],
                   [0, 0, 0])
p.addUserDebugText("Spherical	Drive",
                   [startLocations[1][0], startLocations[1][1] + 1, startLocations[1][2]],
                   [0, 0, 0])
p.addUserDebugText("Explicit PD",
                   [startLocations[2][0], startLocations[2][1] + 1, startLocations[2][2]],
                   [0, 0, 0])
p.addUserDebugText("Kinematic",
                   [startLocations[3][0], startLocations[3][1] + 1, startLocations[3][2]],
                   [0, 0, 0])
p.addUserDebugText("Stable PD (Py)",
                   [startLocations[4][0], startLocations[4][1] + 1, startLocations[4][2]],
                   [0, 0, 0])
flags=p.URDF_MAINTAIN_LINK_ORDER+p.URDF_USE_SELF_COLLISION
humanoid = p.loadURDF("humanoid/humanoid.urdf",
                      startLocations[0],
                      globalScaling=0.25,
                      useFixedBase=False,
                      flags=flags)
humanoid2 = p.loadURDF("humanoid/humanoid.urdf",
                       startLocations[1],
                       globalScaling=0.25,
                       useFixedBase=False,
                       flags=flags)
humanoid3 = p.loadURDF("humanoid/humanoid.urdf",
                       startLocations[2],
                       globalScaling=0.25,
                       useFixedBase=False,
                       flags=flags)
humanoid4 = p.loadURDF("humanoid/humanoid.urdf",
                       startLocations[3],
                       globalScaling=0.25,
                       useFixedBase=False,
                       flags=flags)
humanoid5 = p.loadURDF("humanoid/humanoid.urdf",
                      startLocations[4],
                      globalScaling=0.25,
                      useFixedBase=False,
                      flags=flags)

humanoid_fix = p.createConstraint(humanoid, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0],
                                  startLocations[0], [0, 0, 0, 1])
humanoid2_fix = p.createConstraint(humanoid2, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0],
                                   startLocations[1], [0, 0, 0, 1])
humanoid3_fix = p.createConstraint(humanoid3, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0],
                                   startLocations[2], [0, 0, 0, 1])
humanoid3_fix = p.createConstraint(humanoid4, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0],
                                   startLocations[3], [0, 0, 0, 1])
humanoid4_fix = p.createConstraint(humanoid5, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0],
                                   startLocations[4], [0, 0, 0, 1])

startPose = [
    2, 0.847532, 0, 0.9986781045, 0.01410400148, -0.0006980000731, -0.04942300517, 0.9988133229,
    0.009485003066, -0.04756001538, -0.004475001447, 1, 0, 0, 0, 0.9649395871, 0.02436898957,
    -0.05755497537, 0.2549218909, -0.249116, 0.9993661511, 0.009952001505, 0.03265400494,
    0.01009800153, 0.9854981188, -0.06440700776, 0.09324301124, -0.1262970152, 0.170571,
    0.9927545808, -0.02090099117, 0.08882396249, -0.07817796699, -0.391532, 0.9828788495,
    0.1013909845, -0.05515999155, 0.143618978, 0.9659421276, 0.1884590249, -0.1422460188,
    0.105854014, 0.581348
]

startVel = [
    1.235314324, -0.008525509087, 0.1515293946, -1.161516553, 0.1866449799, -0.1050802848, 0,
    0.935706195, 0.08277326387, 0.3002461862, 0, 0, 0, 0, 0, 1.114409628, 0.3618553952,
    -0.4505575061, 0, -1.725374735, -0.5052852598, -0.8555179722, -0.2221173515, 0, -0.1837617357,
    0.00171895706, 0.03912837591, 0, 0.147945294, 1.837653345, 0.1534535548, 1.491385941, 0,
    -4.632454387, -0.9111172777, -1.300648184, -1.345694622, 0, -1.084238535, 0.1313680236,
    -0.7236998534, 0, -0.5278312973
]

p.resetBasePositionAndOrientation(humanoid, startLocations[0], [0, 0, 0, 1])
p.resetBasePositionAndOrientation(humanoid2, startLocations[1], [0, 0, 0, 1])
p.resetBasePositionAndOrientation(humanoid3, startLocations[2], [0, 0, 0, 1])
p.resetBasePositionAndOrientation(humanoid4, startLocations[3], [0, 0, 0, 1])
p.resetBasePositionAndOrientation(humanoid5, startLocations[4], [0, 0, 0, 1])

index0 = 7
for j in range(p.getNumJoints(humanoid)):
  ji = p.getJointInfo(humanoid, j)
  targetPosition = [0]
  jointType = ji[2]
  if (jointType == p.JOINT_SPHERICAL):
    targetPosition = [
        startPose[index0 + 1], startPose[index0 + 2], startPose[index0 + 3], startPose[index0 + 0]
    ]
    targetVel = [startVel[index0 + 0], startVel[index0 + 1], startVel[index0 + 2]]
    index0 += 4
    print("spherical position: ", targetPosition)
    print("spherical velocity: ", targetVel)
    p.resetJointStateMultiDof(humanoid, j, targetValue=targetPosition, targetVelocity=targetVel)
    p.resetJointStateMultiDof(humanoid5, j, targetValue=targetPosition, targetVelocity=targetVel)
    p.resetJointStateMultiDof(humanoid2, j, targetValue=targetPosition, targetVelocity=targetVel)
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    targetPosition = [startPose[index0]]
    targetVel = [startVel[index0]]
    index0 += 1
    print("revolute:", targetPosition)
    print("revolute	velocity:", targetVel)
    p.resetJointStateMultiDof(humanoid, j, targetValue=targetPosition, targetVelocity=targetVel)
    p.resetJointStateMultiDof(humanoid5, j, targetValue=targetPosition, targetVelocity=targetVel)
    p.resetJointStateMultiDof(humanoid2, j, targetValue=targetPosition, targetVelocity=targetVel)

for j in range(p.getNumJoints(humanoid)):
  ji = p.getJointInfo(humanoid, j)
  targetPosition = [0]
  jointType = ji[2]
  if (jointType == p.JOINT_SPHERICAL):
    targetPosition = [0, 0, 0, 1]
    p.setJointMotorControlMultiDof(humanoid,
                                   j,
                                   p.POSITION_CONTROL,
                                   targetPosition,
                                   targetVelocity=[0, 0, 0],
                                   positionGain=0,
                                   velocityGain=1,
                                   force=[0, 0, 0])
    p.setJointMotorControlMultiDof(humanoid5,
                                   j,
                                   p.POSITION_CONTROL,
                                   targetPosition,
                                   targetVelocity=[0, 0, 0],
                                   positionGain=0,
                                   velocityGain=1,
                                   force=[0, 0, 0])
    p.setJointMotorControlMultiDof(humanoid3,
                                   j,
                                   p.POSITION_CONTROL,
                                   targetPosition,
                                   targetVelocity=[0, 0, 0],
                                   positionGain=0,
                                   velocityGain=1,
                                   force=[31, 31, 31])
    p.setJointMotorControlMultiDof(humanoid4,
                                   j,
                                   p.POSITION_CONTROL,
                                   targetPosition,
                                   targetVelocity=[0, 0, 0],
                                   positionGain=0,
                                   velocityGain=1,
                                   force=[1, 1, 1])

  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    p.setJointMotorControl2(humanoid, j, p.VELOCITY_CONTROL, targetVelocity=0, force=0)
    p.setJointMotorControl2(humanoid3, j, p.VELOCITY_CONTROL, targetVelocity=0, force=31)
    p.setJointMotorControl2(humanoid4, j, p.VELOCITY_CONTROL, targetVelocity=0, force=10)
    p.setJointMotorControl2(humanoid5, j, p.VELOCITY_CONTROL, targetVelocity=0, force=0)

  #print(ji)
  print("joint[", j, "].type=", jointTypes[ji[2]])
  print("joint[", j, "].name=", ji[1])

jointIds = []
paramIds = []
for j in range(p.getNumJoints(humanoid)):
  #p.changeDynamics(humanoid,j,linearDamping=0,	angularDamping=0)
  p.changeVisualShape(humanoid, j, rgbaColor=[1, 1, 1, 1])
  info = p.getJointInfo(humanoid, j)
  #print(info)
  if (not useMotionCapture):
    jointName = info[1]
    jointType = info[2]
    if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
      jointIds.append(j)
      #paramIds.append(p.addUserDebugParameter(jointName.decode("utf-8"),-4,4,0))
      #print("jointName=",jointName, "at ",	j)

p.changeVisualShape(humanoid, 2, rgbaColor=[1, 0, 0, 1])
chest = 1
neck = 2
rightHip = 3
rightKnee = 4
rightAnkle = 5
rightShoulder = 6
rightElbow = 7
leftHip = 9
leftKnee = 10
leftAnkle = 11
leftShoulder = 12
leftElbow = 13

#rightShoulder=3
#rightElbow=4
#leftShoulder=6
#leftElbow = 7
#rightHip	=	9
#rightKnee=10
#rightAnkle=11
#leftHip = 12
#leftKnee=13
#leftAnkle=14

import time

kpOrg = [
    0, 0, 0, 0, 0, 0, 0, 1000, 1000, 1000, 1000, 100, 100, 100, 100, 500, 500, 500, 500, 500, 400,
    400, 400, 400, 400, 400, 400, 400, 300, 500, 500, 500, 500, 500, 400, 400, 400, 400, 400, 400,
    400, 400, 300
]
kdOrg = [
    0, 0, 0, 0, 0, 0, 0, 100, 100, 100, 100, 10, 10, 10, 10, 50, 50, 50, 50, 50, 40, 40, 40, 40,
    40, 40, 40, 40, 30, 50, 50, 50, 50, 50, 40, 40, 40, 40, 40, 40, 40, 40, 30
]

once = True
p.getCameraImage(320, 200)

while (p.isConnected()):

  if useGUI:
    erp = p.readUserDebugParameter(erpId)
    kpMotor = p.readUserDebugParameter(kpMotorId)
    maxForce = p.readUserDebugParameter(forceMotorId)
    frameReal = p.readUserDebugParameter(frameId)
  else:
    erp = 0.2
    kpMotor = 0.2
    maxForce = 1000
    frameReal = 0

  kp = kpMotor

  frame = int(frameReal)
  frameNext = frame + 1
  if (frameNext >= numFrames):
    frameNext = frame

  frameFraction = frameReal - frame
  #print("frameFraction=",frameFraction)
  #print("frame=",frame)
  #print("frameNext=", frameNext)

  #getQuaternionSlerp

  frameData = motion_dict['Frames'][frame]
  frameDataNext = motion_dict['Frames'][frameNext]

  #print("duration=",frameData[0])
  #print(pos=[frameData])

  basePos1Start = [frameData[1], frameData[2], frameData[3]]
  basePos1End = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
  basePos1 = [
      basePos1Start[0] + frameFraction * (basePos1End[0] - basePos1Start[0]),
      basePos1Start[1] + frameFraction * (basePos1End[1] - basePos1Start[1]),
      basePos1Start[2] + frameFraction * (basePos1End[2] - basePos1Start[2])
  ]
  baseOrn1Start = [frameData[5], frameData[6], frameData[7], frameData[4]]
  baseOrn1Next = [frameDataNext[5], frameDataNext[6], frameDataNext[7], frameDataNext[4]]
  baseOrn1 = p.getQuaternionSlerp(baseOrn1Start, baseOrn1Next, frameFraction)
  #pre-rotate	to make	z-up
  if (useZUp):
    y2zPos = [0, 0, 0.0]
    y2zOrn = p.getQuaternionFromEuler([1.57, 0, 0])
    basePos, baseOrn = p.multiplyTransforms(y2zPos, y2zOrn, basePos1, baseOrn1)
    p.resetBasePositionAndOrientation(humanoid, basePos, baseOrn)

    y2zPos = [0, 2, 0.0]
    y2zOrn = p.getQuaternionFromEuler([1.57, 0, 0])
    basePos, baseOrn = p.multiplyTransforms(y2zPos, y2zOrn, basePos1, baseOrn1)
    p.resetBasePositionAndOrientation(humanoid2, basePos, baseOrn)

  chestRotStart = [frameData[9], frameData[10], frameData[11], frameData[8]]
  chestRotEnd = [frameDataNext[9], frameDataNext[10], frameDataNext[11], frameDataNext[8]]
  chestRot = p.getQuaternionSlerp(chestRotStart, chestRotEnd, frameFraction)

  neckRotStart = [frameData[13], frameData[14], frameData[15], frameData[12]]
  neckRotEnd = [frameDataNext[13], frameDataNext[14], frameDataNext[15], frameDataNext[12]]
  neckRot = p.getQuaternionSlerp(neckRotStart, neckRotEnd, frameFraction)

  rightHipRotStart = [frameData[17], frameData[18], frameData[19], frameData[16]]
  rightHipRotEnd = [frameDataNext[17], frameDataNext[18], frameDataNext[19], frameDataNext[16]]
  rightHipRot = p.getQuaternionSlerp(rightHipRotStart, rightHipRotEnd, frameFraction)

  rightKneeRotStart = [frameData[20]]
  rightKneeRotEnd = [frameDataNext[20]]
  rightKneeRot = [
      rightKneeRotStart[0] + frameFraction * (rightKneeRotEnd[0] - rightKneeRotStart[0])
  ]

  rightAnkleRotStart = [frameData[22], frameData[23], frameData[24], frameData[21]]
  rightAnkleRotEnd = [frameDataNext[22], frameDataNext[23], frameDataNext[24], frameDataNext[21]]
  rightAnkleRot = p.getQuaternionSlerp(rightAnkleRotStart, rightAnkleRotEnd, frameFraction)

  rightShoulderRotStart = [frameData[26], frameData[27], frameData[28], frameData[25]]
  rightShoulderRotEnd = [
      frameDataNext[26], frameDataNext[27], frameDataNext[28], frameDataNext[25]
  ]
  rightShoulderRot = p.getQuaternionSlerp(rightShoulderRotStart, rightShoulderRotEnd,
                                          frameFraction)

  rightElbowRotStart = [frameData[29]]
  rightElbowRotEnd = [frameDataNext[29]]
  rightElbowRot = [
      rightElbowRotStart[0] + frameFraction * (rightElbowRotEnd[0] - rightElbowRotStart[0])
  ]

  leftHipRotStart = [frameData[31], frameData[32], frameData[33], frameData[30]]
  leftHipRotEnd = [frameDataNext[31], frameDataNext[32], frameDataNext[33], frameDataNext[30]]
  leftHipRot = p.getQuaternionSlerp(leftHipRotStart, leftHipRotEnd, frameFraction)

  leftKneeRotStart = [frameData[34]]
  leftKneeRotEnd = [frameDataNext[34]]
  leftKneeRot = [leftKneeRotStart[0] + frameFraction * (leftKneeRotEnd[0] - leftKneeRotStart[0])]

  leftAnkleRotStart = [frameData[36], frameData[37], frameData[38], frameData[35]]
  leftAnkleRotEnd = [frameDataNext[36], frameDataNext[37], frameDataNext[38], frameDataNext[35]]
  leftAnkleRot = p.getQuaternionSlerp(leftAnkleRotStart, leftAnkleRotEnd, frameFraction)

  leftShoulderRotStart = [frameData[40], frameData[41], frameData[42], frameData[39]]
  leftShoulderRotEnd = [frameDataNext[40], frameDataNext[41], frameDataNext[42], frameDataNext[39]]
  leftShoulderRot = p.getQuaternionSlerp(leftShoulderRotStart, leftShoulderRotEnd, frameFraction)
  leftElbowRotStart = [frameData[43]]
  leftElbowRotEnd = [frameDataNext[43]]
  leftElbowRot = [
      leftElbowRotStart[0] + frameFraction * (leftElbowRotEnd[0] - leftElbowRotStart[0])
  ]

  if (0):  #if (once):
    p.resetJointStateMultiDof(humanoid, chest, chestRot)
    p.resetJointStateMultiDof(humanoid, neck, neckRot)
    p.resetJointStateMultiDof(humanoid, rightHip, rightHipRot)
    p.resetJointStateMultiDof(humanoid, rightKnee, rightKneeRot)
    p.resetJointStateMultiDof(humanoid, rightAnkle, rightAnkleRot)
    p.resetJointStateMultiDof(humanoid, rightShoulder, rightShoulderRot)
    p.resetJointStateMultiDof(humanoid, rightElbow, rightElbowRot)
    p.resetJointStateMultiDof(humanoid, leftHip, leftHipRot)
    p.resetJointStateMultiDof(humanoid, leftKnee, leftKneeRot)
    p.resetJointStateMultiDof(humanoid, leftAnkle, leftAnkleRot)
    p.resetJointStateMultiDof(humanoid, leftShoulder, leftShoulderRot)
    p.resetJointStateMultiDof(humanoid, leftElbow, leftElbowRot)
  once = False
  #print("chestRot=",chestRot)
  p.setGravity(0, 0, -10)

  kp = kpMotor
  if (useExplicitPD):
    jointDofCounts = [4, 4, 4, 1, 4, 4, 1, 4, 1, 4, 4, 1]
    #[x,y,z] base	position and [x,y,z,w] base	orientation!
    totalDofs = 7
    for dof in jointDofCounts:
      totalDofs += dof

    jointIndicesAll = [
        chest, neck, rightHip, rightKnee, rightAnkle, rightShoulder, rightElbow, leftHip, leftKnee,
        leftAnkle, leftShoulder, leftElbow
    ]
    basePos, baseOrn = p.getBasePositionAndOrientation(humanoid)
    pose = [
        basePos[0], basePos[1], basePos[2], baseOrn[0], baseOrn[1], baseOrn[2], baseOrn[3],
        chestRot[0], chestRot[1], chestRot[2], chestRot[3], neckRot[0], neckRot[1], neckRot[2],
        neckRot[3], rightHipRot[0], rightHipRot[1], rightHipRot[2], rightHipRot[3],
        rightKneeRot[0], rightAnkleRot[0], rightAnkleRot[1], rightAnkleRot[2], rightAnkleRot[3],
        rightShoulderRot[0], rightShoulderRot[1], rightShoulderRot[2], rightShoulderRot[3],
        rightElbowRot[0], leftHipRot[0], leftHipRot[1], leftHipRot[2], leftHipRot[3],
        leftKneeRot[0], leftAnkleRot[0], leftAnkleRot[1], leftAnkleRot[2], leftAnkleRot[3],
        leftShoulderRot[0], leftShoulderRot[1], leftShoulderRot[2], leftShoulderRot[3],
        leftElbowRot[0]
    ]

    #print("pose=")
    #for po	in pose:
    #	print(po)
    
    
    taus = stablePD.computePD(bodyUniqueId=humanoid5,
                                jointIndices=jointIndicesAll,
                                desiredPositions=pose,
                                desiredVelocities=[0] * totalDofs,
                                kps=kpOrg,
                                kds=kdOrg,
                                maxForces=[maxForce] * totalDofs,
                                timeStep=timeStep)
    
    indices = [chest, neck, rightHip, rightKnee,
              rightAnkle, rightShoulder, rightElbow,
               leftHip, leftKnee, leftAnkle,
               leftShoulder, leftElbow]
    targetPositions = [chestRot,neckRot,rightHipRot, rightKneeRot,
                      rightAnkleRot, rightShoulderRot, rightElbowRot,
                      leftHipRot, leftKneeRot, leftAnkleRot,
                      leftShoulderRot, leftElbowRot]
    maxForces = [  [maxForce,maxForce,maxForce], [maxForce,maxForce,maxForce],[maxForce,maxForce,maxForce],[maxForce],
                  [maxForce,maxForce,maxForce],[maxForce,maxForce,maxForce],[maxForce],
                  [maxForce,maxForce,maxForce], [maxForce], [maxForce,maxForce,maxForce],
                  [maxForce,maxForce,maxForce], [maxForce]]
    
   
    kps = [1000]*12
    kds = [100]*12
    
                      
    p.setJointMotorControlMultiDofArray(humanoid,
                                   indices,
                                   p.STABLE_PD_CONTROL,
                                   targetPositions=targetPositions,
                                   positionGains=kps,
                                   velocityGains=kds,
                                   forces=maxForces)

    taus3 = explicitPD.computePD(bodyUniqueId=humanoid3,
                                 jointIndices=jointIndicesAll,
                                 desiredPositions=pose,
                                 desiredVelocities=[0] * totalDofs,
                                 kps=kpOrg,
                                 kds=kdOrg,
                                 maxForces=[maxForce * 0.05] * totalDofs,
                                 timeStep=timeStep)

    #taus=[0]*43
    dofIndex = 7
    for index in range(len(jointIndicesAll)):
      jointIndex = jointIndicesAll[index]
      if jointDofCounts[index] == 4:
        
        p.setJointMotorControlMultiDof(
              humanoid5,
            jointIndex,
            p.TORQUE_CONTROL,
            force=[taus[dofIndex + 0], taus[dofIndex + 1], taus[dofIndex + 2]])
        p.setJointMotorControlMultiDof(
            humanoid3,
            jointIndex,
            p.TORQUE_CONTROL,
            force=[taus3[dofIndex + 0], taus3[dofIndex + 1], taus3[dofIndex + 2]])

      if jointDofCounts[index] == 1:

        
        p.setJointMotorControlMultiDof(humanoid5,
                                       jointIndex,
                                       controlMode=p.TORQUE_CONTROL,
                                       force=[taus[dofIndex]])
        p.setJointMotorControlMultiDof(humanoid3,
                                       jointIndex,
                                       controlMode=p.TORQUE_CONTROL,
                                       force=[taus3[dofIndex]])

      dofIndex += jointDofCounts[index]

    #print("len(taus)=",len(taus))
    #print("taus=",taus)

  p.setJointMotorControlMultiDof(humanoid2,
                                 chest,
                                 p.POSITION_CONTROL,
                                 targetPosition=chestRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 neck,
                                 p.POSITION_CONTROL,
                                 targetPosition=neckRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 rightHip,
                                 p.POSITION_CONTROL,
                                 targetPosition=rightHipRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 rightKnee,
                                 p.POSITION_CONTROL,
                                 targetPosition=rightKneeRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 rightAnkle,
                                 p.POSITION_CONTROL,
                                 targetPosition=rightAnkleRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 rightShoulder,
                                 p.POSITION_CONTROL,
                                 targetPosition=rightShoulderRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 rightElbow,
                                 p.POSITION_CONTROL,
                                 targetPosition=rightElbowRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 leftHip,
                                 p.POSITION_CONTROL,
                                 targetPosition=leftHipRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 leftKnee,
                                 p.POSITION_CONTROL,
                                 targetPosition=leftKneeRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 leftAnkle,
                                 p.POSITION_CONTROL,
                                 targetPosition=leftAnkleRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 leftShoulder,
                                 p.POSITION_CONTROL,
                                 targetPosition=leftShoulderRot,
                                 positionGain=kp,
                                 force=[maxForce])
  p.setJointMotorControlMultiDof(humanoid2,
                                 leftElbow,
                                 p.POSITION_CONTROL,
                                 targetPosition=leftElbowRot,
                                 positionGain=kp,
                                 force=[maxForce])

  kinematicHumanoid4 = True
  if (kinematicHumanoid4):
    p.resetJointStateMultiDof(humanoid4, chest, chestRot)
    p.resetJointStateMultiDof(humanoid4, neck, neckRot)
    p.resetJointStateMultiDof(humanoid4, rightHip, rightHipRot)
    p.resetJointStateMultiDof(humanoid4, rightKnee, rightKneeRot)
    p.resetJointStateMultiDof(humanoid4, rightAnkle, rightAnkleRot)
    p.resetJointStateMultiDof(humanoid4, rightShoulder, rightShoulderRot)
    p.resetJointStateMultiDof(humanoid4, rightElbow, rightElbowRot)
    p.resetJointStateMultiDof(humanoid4, leftHip, leftHipRot)
    p.resetJointStateMultiDof(humanoid4, leftKnee, leftKneeRot)
    p.resetJointStateMultiDof(humanoid4, leftAnkle, leftAnkleRot)
    p.resetJointStateMultiDof(humanoid4, leftShoulder, leftShoulderRot)
    p.resetJointStateMultiDof(humanoid4, leftElbow, leftElbowRot)
  p.stepSimulation()

  if showJointMotorTorques:
    for j in range(p.getNumJoints(humanoid2)):
      jointState = p.getJointStateMultiDof(humanoid2, j)
      print("jointStateMultiDof[", j, "].pos=", jointState[0])
      print("jointStateMultiDof[", j, "].vel=", jointState[1])
      print("jointStateMultiDof[", j, "].jointForces=", jointState[3])
  time.sleep(timeStep)



# Code adapted from https://github.com/DLR-RM/rl-baselines3-zoo
# it requires stable-baselines3 to be installed
# Colab Notebook: https://colab.research.google.com/github/Stable-Baselines-Team/rl-colab-notebooks/blob/sb3/pybullet.ipynb
# You can run it using: python -m pybullet_envs.stable_baselines.enjoy --algo td3 --env HalfCheetahBulletEnv-v0
# Author: Antonin RAFFIN
# MIT License
import os
import time
import argparse

import gym
import numpy as np
import pybullet_envs

from stable_baselines3 import SAC, TD3


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Enjoy an RL agent trained using Stable Baselines3"
    )
    parser.add_argument(
        "--algo",
        help="RL Algorithm (Soft Actor-Critic by default)",
        default="sac",
        type=str,
        required=False,
        choices=["sac", "td3"],
    )
    parser.add_argument(
        "--env", type=str, default="HalfCheetahBulletEnv-v0", help="environment ID"
    )
    parser.add_argument(
        "-n", "--n-episodes", help="Number of episodes", default=5, type=int
    )
    parser.add_argument(
        "--no-render",
        action="store_true",
        default=False,
        help="Do not render the environment",
    )
    parser.add_argument(
        "--load-best",
        action="store_true",
        default=False,
        help="Load best model instead of last model if available",
    )
    args = parser.parse_args()

    env_id = args.env
    # Create an env similar to the training env
    env = gym.make(env_id)

    # Enable GUI
    if not args.no_render:
        env.render(mode="human")

    algo = {
        "sac": SAC,
        "td3": TD3,
    }[args.algo]

    # We assume that the saved model is in the same folder
    save_path = f"{args.algo}_{env_id}.zip"

    if not os.path.isfile(save_path) or args.load_best:
        print("Loading best model")
        # Try to load best model
        save_path = os.path.join(f"{args.algo}_{env_id}", "best_model.zip")

    # Load the saved model
    model = algo.load(save_path, env=env)

    try:
        # Use deterministic actions for evaluation
        episode_rewards, episode_lengths = [], []
        for _ in range(args.n_episodes):
            obs = env.reset()
            done = False
            episode_reward = 0.0
            episode_length = 0
            while not done:
                action, _ = model.predict(obs, deterministic=True)
                obs, reward, done, _info = env.step(action)
                episode_reward += reward

                episode_length += 1
                if not args.no_render:
                    env.render(mode="human")
                    dt = 1.0 / 240.0
                    time.sleep(dt)
            episode_rewards.append(episode_reward)
            episode_lengths.append(episode_length)
            print(
                f"Episode {len(episode_rewards)} reward={episode_reward}, length={episode_length}"
            )

        mean_reward = np.mean(episode_rewards)
        std_reward = np.std(episode_rewards)

        mean_len, std_len = np.mean(episode_lengths), np.std(episode_lengths)

        print("==== Results ====")
        print(f"Episode_reward={mean_reward:.2f} +/- {std_reward:.2f}")
        print(f"Episode_length={mean_len:.2f} +/- {std_len:.2f}")
    except KeyboardInterrupt:
        pass

    # Close process
    env.close()



"""Runs a random policy for the random object KukaDiverseObjectEnv.
"""

import os, inspect
import numpy as np

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import gym
from pybullet_envs.bullet.kuka_diverse_object_gym_env import KukaDiverseObjectEnv
from gym import spaces


class ContinuousDownwardBiasPolicy(object):
  """Policy which takes continuous actions, and is biased to move down.
  """

  def __init__(self, height_hack_prob=0.9):
    """Initializes the DownwardBiasPolicy.

    Args:
        height_hack_prob: The probability of moving down at every move.
    """
    self._height_hack_prob = height_hack_prob
    self._action_space = spaces.Box(low=-1, high=1, shape=(5,))

  def sample_action(self, obs, explore_prob):
    """Implements height hack and grasping threshold hack.
    """
    dx, dy, dz, da, close = self._action_space.sample()
    if np.random.random() < self._height_hack_prob:
      dz = -1
    return [dx, dy, dz, da, 0]


def main():

  env = KukaDiverseObjectEnv(renders=True, isDiscrete=False)
  policy = ContinuousDownwardBiasPolicy()

  while True:
    obs, done = env.reset(), False
    print("===================================")
    print("obs")
    print(obs)
    episode_rew = 0
    while not done:
      env.render(mode='human')
      act = policy.sample_action(obs, .1)
      print("Action")
      print(act)
      obs, rew, done, _ = env.step([0, 0, 0, 0, 0])
      episode_rew += rew
    print("Episode reward", episode_rew)


if __name__ == '__main__':
  main()



# Lint as: python3
r"""An example that the Laikago walks forward using a static gait.

blaze run -c opt //robotics/reinforcement_learning/minitaur/envs_v2/examples:\
laikago_static_gait_example
"""
from absl import app
from absl import flags
import gin
from pybullet_envs.minitaur.agents.baseline_controller import static_gait_controller
from pybullet_envs.minitaur.envs_v2 import env_loader
import pybullet_data as pd

flags.DEFINE_bool("render", True, "Whether to render the example.")

FLAGS = flags.FLAGS
_CONFIG_FILE = pd.getDataPath()+"/configs/laikago_walk_static_gait.gin"
_NUM_STEPS = 10000
_ENV_RANDOM_SEED = 13


def _load_config(render=False):
  gin.parse_config_file(_CONFIG_FILE)
  gin.bind_parameter("SimulationParameters.enable_rendering", render)


def run_example(num_max_steps=_NUM_STEPS):
  """Runs the example.

  Args:
    num_max_steps: Maximum number of steps this example should run for.
  """
  env = env_loader.load()

  env.seed(_ENV_RANDOM_SEED)
  observation = env.reset()
  policy = static_gait_controller.StaticGaitController(env.robot)

  for _ in range(num_max_steps):
    action = policy.act(observation)
    _, _, done, _ = env.step(action)
    if done:
      break


def main(_):
  _load_config(FLAGS.render)
  run_example()


if __name__ == "__main__":
  app.run(main)



"""Motor model for laikago."""

import collections
import numpy as np

from pybullet_envs.minitaur.robots import robot_config

NUM_MOTORS = 12

MOTOR_COMMAND_DIMENSION = 5

# These values represent the indices of each field in the motor command tuple
POSITION_INDEX = 0
POSITION_GAIN_INDEX = 1
VELOCITY_INDEX = 2
VELOCITY_GAIN_INDEX = 3
TORQUE_INDEX = 4


class LaikagoMotorModel(object):
  """A simple motor model for Laikago.

    When in POSITION mode, the torque is calculated according to the difference
    between current and desired joint angle, as well as the joint velocity.
    For more information about PD control, please refer to:
    https://en.wikipedia.org/wiki/PID_controller.

    The model supports a HYBRID mode in which each motor command can be a tuple
    (desired_motor_angle, position_gain, desired_motor_velocity, velocity_gain,
    torque).

  """

  def __init__(self,
               kp=60,
               kd=1,
               torque_limits=None,
               motor_control_mode=robot_config.MotorControlMode.POSITION):
    self._kp = kp
    self._kd = kd
    self._torque_limits = torque_limits
    if torque_limits is not None:
      if isinstance(torque_limits, (collections.Sequence, np.ndarray)):
        self._torque_limits = np.asarray(torque_limits)
      else:
        self._torque_limits = np.full(NUM_MOTORS, torque_limits)
    self._motor_control_mode = motor_control_mode
    self._strength_ratios = np.full(NUM_MOTORS, 1)

  def set_strength_ratios(self, ratios):
    """Set the strength of each motors relative to the default value.

    Args:
      ratios: The relative strength of motor output. A numpy array ranging from
        0.0 to 1.0.
    """
    self._strength_ratios = ratios

  def set_motor_gains(self, kp, kd):
    """Set the gains of all motors.

    These gains are PD gains for motor positional control. kp is the
    proportional gain and kd is the derivative gain.

    Args:
      kp: proportional gain of the motors.
      kd: derivative gain of the motors.
    """
    self._kp = kp
    self._kd = kd

  def set_voltage(self, voltage):
    pass

  def get_voltage(self):
    return 0.0

  def set_viscous_damping(self, viscous_damping):
    pass

  def get_viscous_dampling(self):
    return 0.0

  def convert_to_torque(self,
                        motor_commands,
                        motor_angle,
                        motor_velocity,
                        true_motor_velocity,
                        motor_control_mode=None):
    """Convert the commands (position control or torque control) to torque.

    Args:
      motor_commands: The desired motor angle if the motor is in position
        control mode. The pwm signal if the motor is in torque control mode.
      motor_angle: The motor angle observed at the current time step. It is
        actually the true motor angle observed a few milliseconds ago (pd
        latency).
      motor_velocity: The motor velocity observed at the current time step, it
        is actually the true motor velocity a few milliseconds ago (pd latency).
      true_motor_velocity: The true motor velocity. The true velocity is used to
        compute back EMF voltage and viscous damping.
      motor_control_mode: A MotorControlMode enum.

    Returns:
      actual_torque: The torque that needs to be applied to the motor.
      observed_torque: The torque observed by the sensor.
    """
    del true_motor_velocity
    if not motor_control_mode:
      motor_control_mode = self._motor_control_mode

    # No processing for motor torques
    if motor_control_mode is robot_config.MotorControlMode.TORQUE:
      assert len(motor_commands) == NUM_MOTORS
      motor_torques = self._strength_ratios * motor_commands
      return motor_torques, motor_torques

    desired_motor_angles = None
    desired_motor_velocities = None
    kp = None
    kd = None
    additional_torques = np.full(NUM_MOTORS, 0)
    if motor_control_mode is robot_config.MotorControlMode.POSITION:
      assert len(motor_commands) == NUM_MOTORS
      kp = self._kp
      kd = self._kd
      desired_motor_angles = motor_commands
      desired_motor_velocities = np.full(NUM_MOTORS, 0)
    elif motor_control_mode is robot_config.MotorControlMode.HYBRID:
      # The input should be a 60 dimension vector
      assert len(motor_commands) == MOTOR_COMMAND_DIMENSION * NUM_MOTORS
      kp = motor_commands[POSITION_GAIN_INDEX::MOTOR_COMMAND_DIMENSION]
      kd = motor_commands[VELOCITY_GAIN_INDEX::MOTOR_COMMAND_DIMENSION]
      desired_motor_angles = motor_commands[
          POSITION_INDEX::MOTOR_COMMAND_DIMENSION]
      desired_motor_velocities = motor_commands[
          VELOCITY_INDEX::MOTOR_COMMAND_DIMENSION]
      additional_torques = motor_commands[TORQUE_INDEX::MOTOR_COMMAND_DIMENSION]
    motor_torques = -1 * (kp * (motor_angle - desired_motor_angles)) - kd * (
        motor_velocity - desired_motor_velocities) + additional_torques
    motor_torques = self._strength_ratios * motor_torques
    if self._torque_limits is not None:
      if len(self._torque_limits) != len(motor_torques):
        raise ValueError(
            "Torque limits dimension does not match the number of motors.")
      motor_torques = np.clip(motor_torques, -1.0 * self._torque_limits,
                              self._torque_limits)

    return motor_torques, motor_torques



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import gym
from pybullet_envs.bullet.kukaGymEnv import KukaGymEnv

from baselines import deepq

import datetime


def callback(lcl, glb):
  # stop training if reward exceeds 199
  total = sum(lcl['episode_rewards'][-101:-1]) / 100
  totalt = lcl['t']
  #print("totalt")
  #print(totalt)
  is_solved = totalt > 2000 and total >= 10
  return is_solved


def main():

  env = KukaGymEnv(renders=False, isDiscrete=True)
  model = deepq.models.mlp([64])
  act = deepq.learn(env,
                    q_func=model,
                    lr=1e-3,
                    max_timesteps=10000000,
                    buffer_size=50000,
                    exploration_fraction=0.1,
                    exploration_final_eps=0.02,
                    print_freq=10,
                    callback=callback)
  print("Saving model to kuka_model.pkl")
  act.save("kuka_model.pkl")


if __name__ == '__main__':
  main()



"""A wrapper for motion imitation environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gin
import gym
import numpy as np


@gin.configurable
class ImitationWrapperEnv(object):
  """An env using for training policy with motion imitation."""

  def __init__(self, gym_env):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
    """
    self._gym_env = gym_env
    self.observation_space = self._build_observation_space()

    self.seed()
    return

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the modified observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.

    """
    original_observation, reward, done, _ = self._gym_env.step(action)
    observation = self._modify_observation(original_observation)

    return observation, reward, done, _

  @gin.configurable('imitation_wrapper_env.ImitationWrapperEnv.reset')
  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    """Resets the robot's position in the world or rebuild the sim world.

    The simulation world will be rebuilt if self._hard_reset is True.

    Args:
      initial_motor_angles: A list of Floats. The desired joint angles after
        reset. If None, the robot will use its built-in value.
      reset_duration: Float. The time (in seconds) needed to rotate all motors
        to the desired initial values.

    Returns:
      A numpy array contains the initial observation after reset.
    """
    original_observation = self._gym_env.reset(initial_motor_angles, reset_duration)
    observation = self._modify_observation(original_observation)
    return observation

  def _modify_observation(self, original_observation):
    """Appends target observations from the reference motion to the observations.

    Args:
      original_observation: A numpy array containing the original observations.

    Returns:
      A numpy array contains the initial original concatenated with target
      observations from the reference motion.
    """
    target_observation = self._task.build_target_obs()
    observation = np.concatenate([original_observation, target_observation], axis=-1)
    return observation

  def _build_observation_space(self):
    """Constructs the observation space, including target observations from
    the reference motion.

    Returns:
      Observation space representing the concatenations of the original
      observations and target observations.
    """
    obs_space0 = self._gym_env.observation_space
    low0 = obs_space0.low
    high0 = obs_space0.high

    task_low, task_high = self._task.get_target_obs_bounds()
    low = np.concatenate([low0, task_low], axis=-1)
    high = np.concatenate([high0, task_high], axis=-1)

    obs_space = gym.spaces.Box(low, high)

    return obs_space



import pybullet as p
import pybullet_data as pd
import time

p.connect(p.GUI)
p.setGravity(0,0,-9.8)
p.setAdditionalSearchPath(pd.getDataPath())
floor = p.loadURDF("plane.urdf")
startPos = [0,0,0.5]
robot = p.loadURDF("mini_cheetah/mini_cheetah.urdf",startPos)
numJoints = p.getNumJoints(robot)
p.changeVisualShape(robot,-1,rgbaColor=[1,1,1,1])
for j in range (numJoints):
	p.changeVisualShape(robot,j,rgbaColor=[1,1,1,1])
	force=200
	pos=0
	p.setJointMotorControl2(robot,j,p.POSITION_CONTROL,pos,force=force)
dt = 1./240.
p.setTimeStep(dt)
while (1):
  p.stepSimulation()
  time.sleep(dt)





#using the eglRendererPlugin (hardware OpenGL acceleration)
#using EGL on Linux and default OpenGL window on Win32.

#make sure to compile pybullet with PYBULLET_USE_NUMPY enabled
#otherwise use testrender.py (slower but compatible without numpy)
#you can also use GUI mode, for faster OpenGL rendering (instead of TinyRender CPU)

import numpy as np
import matplotlib.pyplot as plt
import pybullet
import time
import pkgutil

plt.ion()

img = np.random.rand(200, 320)
#img = [tandard_normal((50,100))
image = plt.imshow(img, interpolation='none', animated=True, label="blah")
ax = plt.gca()
import pybullet_data


pybullet.connect(pybullet.DIRECT)

pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())

egl = pkgutil.get_loader('eglRenderer')
if (egl):
  pluginId = pybullet.loadPlugin(egl.get_filename(), "_eglRendererPlugin")
else:
  pluginId = pybullet.loadPlugin("eglRendererPlugin")
print("pluginId=",pluginId)
pybullet.loadURDF("plane.urdf", [0, 0, -1])
pybullet.loadURDF("r2d2.urdf")

camTargetPos = [0, 0, 0]
cameraUp = [0, 0, 1]
cameraPos = [1, 1, 1]
pybullet.setGravity(0, 0, -10)

pitch = -10.0

roll = 0
upAxisIndex = 2
camDistance = 4
pixelWidth = 320
pixelHeight = 200
nearPlane = 0.01
farPlane = 100

fov = 60

main_start = time.time()
while (1):
  for yaw in range(0, 360, 10):
    pybullet.stepSimulation()
    start = time.time()

    viewMatrix = pybullet.computeViewMatrixFromYawPitchRoll(camTargetPos, camDistance, yaw, pitch,
                                                            roll, upAxisIndex)
    aspect = pixelWidth / pixelHeight
    projectionMatrix = pybullet.computeProjectionMatrixFOV(fov, aspect, nearPlane, farPlane)
    img_arr = pybullet.getCameraImage(pixelWidth,
                                      pixelHeight,
                                      viewMatrix,
                                      projectionMatrix,
                                      shadow=1,
                                      lightDirection=[1, 1, 1],
                                      renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)
    stop = time.time()
    #print("renderImage %f" % (stop - start))

    w = img_arr[0]  #width of the image, in pixels
    h = img_arr[1]  #height of the image, in pixels
    rgb = img_arr[2]  #color data RGB
    dep = img_arr[3]  #depth data

    #print('width = %d height = %d' % (w, h))

    #note that sending the data to matplotlib is really slow

    #reshape is not needed
    np_img_arr = np.reshape(rgb, (h, w, 4))
    np_img_arr = np_img_arr * (1. / 255.)

    #show
    #plt.imshow(np_img_arr,interpolation='none',extent=(0,1600,0,1200))
    #image = plt.imshow(np_img_arr,interpolation='none',animated=True,label="blah")

    image.set_data(np_img_arr)
    ax.plot([0])
    #plt.draw()
    #plt.show()
    plt.pause(0.01)
    #image.draw()

main_stop = time.time()

print("Total time %f" % (main_stop - main_start))

pybullet.resetSimulation()



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In-graph simulation step of a vectorized algorithm with environments."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from . import streaming_mean


def simulate(batch_env, algo, log=True, reset=False):
  """Simulation step of a vecrotized algorithm with in-graph environments.

  Integrates the operations implemented by the algorithm and the environments
  into a combined operation.

  Args:
    batch_env: In-graph batch environment.
    algo: Algorithm instance implementing required operations.
    log: Tensor indicating whether to compute and return summaries.
    reset: Tensor causing all environments to reset.

  Returns:
    Tuple of tensors containing done flags for the current episodes, possibly
    intermediate scores for the episodes, and a summary tensor.
  """

  def _define_begin_episode(agent_indices):
    """Reset environments, intermediate scores and durations for new episodes.

    Args:
      agent_indices: Tensor containing batch indices starting an episode.

    Returns:
      Summary tensor.
    """
    assert agent_indices.shape.ndims == 1
    zero_scores = tf.zeros_like(agent_indices, tf.float32)
    zero_durations = tf.zeros_like(agent_indices)
    reset_ops = [
        batch_env.reset(agent_indices),
        tf.scatter_update(score, agent_indices, zero_scores),
        tf.scatter_update(length, agent_indices, zero_durations)
    ]
    with tf.control_dependencies(reset_ops):
      return algo.begin_episode(agent_indices)

  def _define_step():
    """Request actions from the algorithm and apply them to the environments.

    Increments the lengths of all episodes and increases their scores by the
    current reward. After stepping the environments, provides the full
    transition tuple to the algorithm.

    Returns:
      Summary tensor.
    """
    prevob = batch_env.observ + 0  # Ensure a copy of the variable value.
    agent_indices = tf.range(len(batch_env))
    action, step_summary = algo.perform(agent_indices, prevob)
    action.set_shape(batch_env.action.shape)
    with tf.control_dependencies([batch_env.simulate(action)]):
      add_score = score.assign_add(batch_env.reward)
      inc_length = length.assign_add(tf.ones(len(batch_env), tf.int32))
    with tf.control_dependencies([add_score, inc_length]):
      agent_indices = tf.range(len(batch_env))
      experience_summary = algo.experience(agent_indices, prevob, batch_env.action,
                                           batch_env.reward, batch_env.done, batch_env.observ)
    return tf.summary.merge([step_summary, experience_summary])

  def _define_end_episode(agent_indices):
    """Notify the algorithm of ending episodes.

    Also updates the mean score and length counters used for summaries.

    Args:
      agent_indices: Tensor holding batch indices that end their episodes.

    Returns:
      Summary tensor.
    """
    assert agent_indices.shape.ndims == 1
    submit_score = mean_score.submit(tf.gather(score, agent_indices))
    submit_length = mean_length.submit(tf.cast(tf.gather(length, agent_indices), tf.float32))
    with tf.control_dependencies([submit_score, submit_length]):
      return algo.end_episode(agent_indices)

  def _define_summaries():
    """Reset the average score and duration, and return them as summary.

    Returns:
      Summary string.
    """
    score_summary = tf.cond(tf.logical_and(log, tf.cast(
        mean_score.count, tf.bool)), lambda: tf.summary.scalar('mean_score', mean_score.clear()),
                            str)
    length_summary = tf.cond(tf.logical_and(
        log, tf.cast(mean_length.count,
                     tf.bool)), lambda: tf.summary.scalar('mean_length', mean_length.clear()), str)
    return tf.summary.merge([score_summary, length_summary])

  with tf.name_scope('simulate'):
    log = tf.convert_to_tensor(log)
    reset = tf.convert_to_tensor(reset)
    with tf.variable_scope('simulate_temporary'):
      score = tf.Variable(tf.zeros(len(batch_env), dtype=tf.float32), False, name='score')
      length = tf.Variable(tf.zeros(len(batch_env), dtype=tf.int32), False, name='length')
    mean_score = streaming_mean.StreamingMean((), tf.float32)
    mean_length = streaming_mean.StreamingMean((), tf.float32)
    agent_indices = tf.cond(reset, lambda: tf.range(len(batch_env)), lambda: tf.cast(
        tf.where(batch_env.done)[:, 0], tf.int32))
    begin_episode = tf.cond(tf.cast(tf.shape(agent_indices)[0],
                                    tf.bool), lambda: _define_begin_episode(agent_indices), str)
    with tf.control_dependencies([begin_episode]):
      step = _define_step()
    with tf.control_dependencies([step]):
      agent_indices = tf.cast(tf.where(batch_env.done)[:, 0], tf.int32)
      end_episode = tf.cond(tf.cast(tf.shape(agent_indices)[0],
                                    tf.bool), lambda: _define_end_episode(agent_indices), str)
    with tf.control_dependencies([end_episode]):
      summary = tf.summary.merge([_define_summaries(), begin_episode, step, end_episode])
    with tf.control_dependencies([summary]):
      done, score = tf.identity(batch_env.done), tf.identity(score)
    return done, score, summary



"""
Classic cart-pole system implemented by Rich Sutton et al.
Copied from http://incompleteideas.net/book/code/pole.c
"""
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import logging
import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import time
import subprocess
import pybullet as p2
import pybullet_data
from pybullet_utils import bullet_client as bc
from pkg_resources import parse_version

logger = logging.getLogger(__name__)


class CartPoleBulletEnv(gym.Env):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False, discrete_actions=True):
    # start the bullet physics server
    self._renders = renders
    self._discrete_actions = discrete_actions
    self._render_height = 200
    self._render_width = 320
    self._physics_client_id = -1
    self.theta_threshold_radians = 12 * 2 * math.pi / 360
    self.x_threshold = 0.4  #2.4
    high = np.array([
        self.x_threshold * 2,
        np.finfo(np.float32).max, self.theta_threshold_radians * 2,
        np.finfo(np.float32).max
    ])

    self.force_mag = 10

    if self._discrete_actions:
      self.action_space = spaces.Discrete(2)
    else:
      action_dim = 1
      action_high = np.array([self.force_mag] * action_dim)
      self.action_space = spaces.Box(-action_high, action_high)
    
    self.observation_space = spaces.Box(-high, high, dtype=np.float32)

    self.seed()
    #    self.reset()
    self.viewer = None
    self._configure()

  def _configure(self, display=None):
    self.display = display

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def step(self, action):
    p = self._p
    if self._discrete_actions:
      force = self.force_mag if action == 1 else -self.force_mag
    else:
      force = action[0]

    p.setJointMotorControl2(self.cartpole, 0, p.TORQUE_CONTROL, force=force)
    p.stepSimulation()

    self.state = p.getJointState(self.cartpole, 1)[0:2] + p.getJointState(self.cartpole, 0)[0:2]
    theta, theta_dot, x, x_dot = self.state

    done =  x < -self.x_threshold \
                or x > self.x_threshold \
                or theta < -self.theta_threshold_radians \
                or theta > self.theta_threshold_radians
    done = bool(done)
    reward = 1.0
    #print("state=",self.state)
    return np.array(self.state), reward, done, {}

  def reset(self):
    #    print("-----------reset simulation---------------")
    if self._physics_client_id < 0:
      if self._renders:
        self._p = bc.BulletClient(connection_mode=p2.GUI)
      else:
        self._p = bc.BulletClient()
      self._physics_client_id = self._p._client
 
      p = self._p
      p.resetSimulation()
      self.cartpole = p.loadURDF(os.path.join(pybullet_data.getDataPath(), "cartpole.urdf"),
                                 [0, 0, 0])
      p.changeDynamics(self.cartpole, -1, linearDamping=0, angularDamping=0)
      p.changeDynamics(self.cartpole, 0, linearDamping=0, angularDamping=0)
      p.changeDynamics(self.cartpole, 1, linearDamping=0, angularDamping=0)
      self.timeStep = 0.02
      p.setJointMotorControl2(self.cartpole, 1, p.VELOCITY_CONTROL, force=0)
      p.setJointMotorControl2(self.cartpole, 0, p.VELOCITY_CONTROL, force=0)
      p.setGravity(0, 0, -9.8)
      p.setTimeStep(self.timeStep)
      p.setRealTimeSimulation(0)
    p = self._p
    randstate = self.np_random.uniform(low=-0.05, high=0.05, size=(4,))
    p.resetJointState(self.cartpole, 1, randstate[0], randstate[1])
    p.resetJointState(self.cartpole, 0, randstate[2], randstate[3])
    #print("randstate=",randstate)
    self.state = p.getJointState(self.cartpole, 1)[0:2] + p.getJointState(self.cartpole, 0)[0:2]
    #print("self.state=", self.state)
    return np.array(self.state)

  def render(self, mode='human', close=False):
    if mode == "human":
      self._renders = True
    if mode != "rgb_array":
      return np.array([])
    base_pos=[0,0,0]
    self._cam_dist = 2
    self._cam_pitch = 0.3
    self._cam_yaw = 0
    if (self._physics_client_id>=0):
      view_matrix = self._p.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._cam_dist,
        yaw=self._cam_yaw,
        pitch=self._cam_pitch,
        roll=0,
        upAxisIndex=2)
      proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
             aspect=float(self._render_width) /
             self._render_height,
             nearVal=0.1,
             farVal=100.0)
      (_, _, px, _, _) = self._p.getCameraImage(
          width=self._render_width,
          height=self._render_height,
          renderer=self._p.ER_BULLET_HARDWARE_OPENGL,
          viewMatrix=view_matrix,
          projectionMatrix=proj_matrix)
    else:
      px = np.array([[[255,255,255,255]]*self._render_width]*self._render_height, dtype=np.uint8)
    rgb_array = np.array(px, dtype=np.uint8)
    rgb_array = np.reshape(np.array(px), (self._render_height, self._render_width, -1))
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def configure(self, args):
    pass
    
  def close(self):
    if self._physics_client_id >= 0:
      self._p.disconnect()
    self._physics_client_id = -1

class CartPoleContinuousBulletEnv(CartPoleBulletEnv):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False):
    # start the bullet physics server
    CartPoleBulletEnv.__init__(self, renders, discrete_actions=False)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the PPO algorithm usage example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import itertools

import tf.compat.v1 as tf

from google3.robotics.reinforcement_learning.agents import ppo
from google3.robotics.reinforcement_learning.agents import tools
from google3.robotics.reinforcement_learning.agents.scripts import configs
from google3.robotics.reinforcement_learning.agents.scripts import networks
from google3.robotics.reinforcement_learning.agents.scripts import train

FLAGS = tf.app.flags.FLAGS


class PPOTest(tf.test.TestCase):

  def test_no_crash_cheetah(self):
    nets = networks.ForwardGaussianPolicy, networks.RecurrentGaussianPolicy
    for network in nets:
      config = self._define_config()
      with config.unlocked:
        config.env = 'HalfCheetah-v1'
        config.max_length = 200
        config.steps = 1000
        config.network = network
      for score in train.train(config, env_processes=True):
        float(score)

  def test_no_crash_ant(self):
    nets = networks.ForwardGaussianPolicy, networks.RecurrentGaussianPolicy
    for network in nets:
      config = self._define_config()
      with config.unlocked:
        config.env = 'Ant-v1'
        config.max_length = 200
        config.steps = 1000
        config.network = network
      for score in train.train(config, env_processes=True):
        float(score)

  def test_no_crash_observation_shape(self):
    nets = networks.ForwardGaussianPolicy, networks.RecurrentGaussianPolicy
    observ_shapes = (1,), (2, 3), (2, 3, 4)
    for network, observ_shape in itertools.product(nets, observ_shapes):
      config = self._define_config()
      with config.unlocked:
        config.env = functools.partial(tools.MockEnvironment,
                                       observ_shape,
                                       action_shape=(3,),
                                       min_duration=15,
                                       max_duration=15)
        config.max_length = 20
        config.steps = 100
        config.network = network
      for score in train.train(config, env_processes=False):
        float(score)

  def test_no_crash_variable_duration(self):
    config = self._define_config()
    with config.unlocked:
      config.env = functools.partial(tools.MockEnvironment,
                                     observ_shape=(2, 3),
                                     action_shape=(3,),
                                     min_duration=5,
                                     max_duration=25)
      config.max_length = 25
      config.steps = 200
      config.network = networks.RecurrentGaussianPolicy
    for score in train.train(config, env_processes=False):
      float(score)

  def _define_config(self):
    # Start from the example configuration.
    locals().update(configs.default())
    # pylint: disable=unused-variable
    # General
    algorithm = ppo.PPOAlgorithm
    num_agents = 2
    update_every = 4
    use_gpu = False
    # Network
    policy_layers = 20, 10
    value_layers = 20, 10
    # Optimization
    update_epochs_policy = 2
    update_epochs_value = 2
    # pylint: enable=unused-variable
    return tools.AttrDict(locals())


if __name__ == '__main__':
  FLAGS.config = 'unused'
  tf.test.main()



"""Utilities to convert imagery protobufs to other formats."""

import numpy as np

from pybullet_envs.minitaur.vision import imagery_pb2


# TODO(b/123306148): Support the conversion from image array to the proto.
def convert_image_to_array(image):
  """Converts an Image proto into a numpy array.

  Args:
    image: An instance of the imagery_pb2.Image proto.

  Returns:
    A numpy array. For color images (e.g. BGRA), the converted ND array
    has the format [Height, Width, Channel]. For gray images (e.g. depth), the
    converted ND array has the format [Height, Width].
  """

  if image.image_format == imagery_pb2.Image.IMAGE_FORMAT_BGRA_HWC_8U:
    img_buffer = np.fromstring(image.content, dtype=np.uint8)
    img = np.reshape(
        img_buffer, [image.height_px, image.width_px, 4], order="C")
    return img

  if image.image_format == imagery_pb2.Image.IMAGE_FORMAT_RGB_HWC_8U:
    img_buffer = np.fromstring(image.content, dtype=np.uint8)
    img = np.reshape(
        img_buffer, [image.height_px, image.width_px, 3], order="C")
    return img

  if image.image_format == imagery_pb2.Image.IMAGE_FORMAT_GRAY_HW_32F:
    img_buffer = np.fromstring(image.content, dtype=np.float32)
    img = np.reshape(img_buffer, [image.height_px, image.width_px], order="C")
    return img

  if image.image_format == imagery_pb2.Image.IMAGE_FORMAT_GRAY_HW_16U:
    img_buffer = np.fromstring(image.content, dtype=np.uint16)
    img = np.reshape(img_buffer, [image.height_px, image.width_px], order="C")
    return img

  raise ValueError("Unsupported image format {}".format(image.image_format))



import pybullet_envs.deep_mimic.learning.nets.fc_2layers_1024units as fc_2layers_1024units


def build_net(net_name, input_tfs, reuse=False):
  net = None

  if (net_name == fc_2layers_1024units.NAME):
    net = fc_2layers_1024units.build_net(input_tfs, reuse)
  else:
    assert False, 'Unsupported net: ' + net_name

  return net



"""This file implements the gym environment of minitaur alternating legs.

"""
import math

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

from gym import spaces
import numpy as np
from pybullet_envs.minitaur.envs import minitaur_gym_env

INIT_EXTENSION_POS = 2.6
INIT_SWING_POS = 0.0
DESIRED_PITCH = 0
NUM_LEGS = 4
NUM_MOTORS = 2 * NUM_LEGS
STEP_PERIOD = 1.0 / 3.0  # Three steps per second.
STEP_AMPLITUDE = 0.75


class MinitaurAlternatingLegsEnv(minitaur_gym_env.MinitaurGymEnv):
  """The gym environment for the minitaur.

  It simulates the locomotion of a minitaur, a quadruped robot. The state space
  include the angles, velocities and torques for all the motors and the action
  space is the desired motor angle for each motor. The reward function is based
  on how far the minitaur walks in 1000 steps and penalizes the energy
  expenditure.

  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 66}

  def __init__(self,
               urdf_version=None,
               control_time_step=0.006,
               action_repeat=6,
               control_latency=0,
               pd_latency=0,
               on_rack=False,
               motor_kp=1.0,
               motor_kd=0.02,
               remove_default_joint_damping=False,
               render=False,
               num_steps_to_log=1000,
               env_randomizer=None,
               log_path=None):
    """Initialize the minitaur alternating legs gym environment.

    Args:
      urdf_version: [DEFAULT_URDF_VERSION, DERPY_V0_URDF_VERSION] are allowable
        versions. If None, DEFAULT_URDF_VERSION is used. Refer to
        minitaur_gym_env for more details.
      control_time_step: The time step between two successive control signals.
      action_repeat: The number of simulation steps that an action is repeated.
      control_latency: The latency between get_observation() and the actual
        observation. See minituar.py for more details.
      pd_latency: The latency used to get motor angles/velocities used to
        compute PD controllers. See minitaur.py for more details.
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hung midair so
        that its walking gait is clearer to visualize.
      motor_kp: The P gain of the motor.
      motor_kd: The D gain of the motor.
      remove_default_joint_damping: Whether to remove the default joint damping.
      render: Whether to render the simulation.
      num_steps_to_log: The max number of control steps in one episode. If the
        number of steps is over num_steps_to_log, the environment will still
        be running, but only first num_steps_to_log will be recorded in logging.
      env_randomizer: An instance (or a list) of EnvRanzomier(s) that can
        randomize the environment during when env.reset() is called and add
        perturbations when env.step() is called.
      log_path: The path to write out logs. For the details of logging, refer to
        minitaur_logging.proto.
    """
    # _swing_offset and _extension_offset is to mimick the bent legs.
    self._swing_offset = np.zeros(NUM_LEGS)
    self._extension_offset = np.zeros(NUM_LEGS)
    super(MinitaurAlternatingLegsEnv,
          self).__init__(urdf_version=urdf_version,
                         accurate_motor_model_enabled=True,
                         motor_overheat_protection=True,
                         hard_reset=False,
                         motor_kp=motor_kp,
                         motor_kd=motor_kd,
                         remove_default_joint_damping=remove_default_joint_damping,
                         control_latency=control_latency,
                         pd_latency=pd_latency,
                         on_rack=on_rack,
                         render=render,
                         num_steps_to_log=num_steps_to_log,
                         env_randomizer=env_randomizer,
                         log_path=log_path,
                         control_time_step=control_time_step,
                         action_repeat=action_repeat)

    action_dim = 8
    action_high = np.array([0.1] * action_dim)
    self.action_space = spaces.Box(-action_high, action_high)

    self._cam_dist = 1.0
    self._cam_yaw = 30
    self._cam_pitch = -30

  def reset(self):
    self.desired_pitch = DESIRED_PITCH
    # In this environment, the actions are
    # [swing leg 1, swing leg 2, swing leg 3, swing leg 4,
    #  extension leg 1, extension leg 2, extension leg 3, extension leg 4]
    init_pose = [
        INIT_SWING_POS + self._swing_offset[0], INIT_SWING_POS + self._swing_offset[1],
        INIT_SWING_POS + self._swing_offset[2], INIT_SWING_POS + self._swing_offset[3],
        INIT_EXTENSION_POS + self._extension_offset[0],
        INIT_EXTENSION_POS + self._extension_offset[1],
        INIT_EXTENSION_POS + self._extension_offset[2],
        INIT_EXTENSION_POS + self._extension_offset[3]
    ]
    initial_motor_angles = self._convert_from_leg_model(init_pose)
    super(MinitaurAlternatingLegsEnv, self).reset(initial_motor_angles=initial_motor_angles,
                                                  reset_duration=0.5)
    return self._get_observation()

  def _convert_from_leg_model(self, leg_pose):
    motor_pose = np.zeros(NUM_MOTORS)
    for i in range(NUM_LEGS):
      motor_pose[2 * i] = leg_pose[NUM_LEGS + i] - (-1)**(i / 2) * leg_pose[i]
      motor_pose[2 * i + 1] = leg_pose[NUM_LEGS + i] + (-1)**(i / 2) * leg_pose[i]
    return motor_pose

  def _signal(self, t):
    initial_pose = np.array([
        INIT_SWING_POS, INIT_SWING_POS, INIT_SWING_POS, INIT_SWING_POS, INIT_EXTENSION_POS,
        INIT_EXTENSION_POS, INIT_EXTENSION_POS, INIT_EXTENSION_POS
    ])
    amplitude = STEP_AMPLITUDE
    period = STEP_PERIOD
    extension = amplitude * (-1.0 + math.cos(2 * math.pi / period * t))
    ith_leg = int(t / period) % 2
    first_leg = np.array([0, 0, 0, 0, 0, extension, extension, 0])
    second_leg = np.array([0, 0, 0, 0, extension, 0, 0, extension])
    if ith_leg:
      signal = initial_pose + second_leg
    else:
      signal = initial_pose + first_leg
    return signal

  def _transform_action_to_motor_command(self, action):
    #  Add swing_offset and extension_offset to mimick the bent legs.
    action[0:4] += self._swing_offset
    action[4:8] += self._extension_offset
    action += self._signal(self.minitaur.GetTimeSinceReset())
    action = self._convert_from_leg_model(action)
    return action

  def is_fallen(self):
    """Decide whether the minitaur has fallen.

    If the up directions between the base and the world is large (the dot
    product is smaller than 0.85), the minitaur is considered fallen.

    Returns:
      Boolean value that indicates whether the minitaur has fallen.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_mat = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_up = rot_mat[6:]
    return np.dot(np.asarray([0, 0, 1]), np.asarray(local_up)) < 0.85

  def _reward(self):
    return 1.0

  def _get_true_observation(self):
    """Get the true observations of this environment.

    It includes the roll, the error between current pitch and desired pitch,
    roll dot and pitch dot of the base.

    Returns:
      The observation list.
    """
    observation = []
    roll, pitch, _ = self.minitaur.GetTrueBaseRollPitchYaw()
    roll_rate, pitch_rate, _ = self.minitaur.GetTrueBaseRollPitchYawRate()
    observation.extend([roll, pitch, roll_rate, pitch_rate])
    observation[1] -= self.desired_pitch  # observation[1] is the pitch
    self._true_observation = np.array(observation)
    return self._true_observation

  def _get_observation(self):
    observation = []
    roll, pitch, _ = self.minitaur.GetBaseRollPitchYaw()
    roll_rate, pitch_rate, _ = self.minitaur.GetBaseRollPitchYawRate()
    observation.extend([roll, pitch, roll_rate, pitch_rate])
    observation[1] -= self.desired_pitch  # observation[1] is the pitch
    self._observation = np.array(observation)
    return self._observation

  def _get_observation_upper_bound(self):
    """Get the upper bound of the observation.

    Returns:
      The upper bound of an observation. See GetObservation() for the details
        of each element of an observation.
    """
    upper_bound = np.zeros(self._get_observation_dimension())
    upper_bound[0:2] = 2 * math.pi  # Roll, pitch, yaw of the base.
    upper_bound[2:4] = 2 * math.pi / self._time_step  # Roll, pitch, yaw rate.
    return upper_bound

  def _get_observation_lower_bound(self):
    lower_bound = -self._get_observation_upper_bound()
    return lower_bound

  def set_swing_offset(self, value):
    """Set the swing offset of each leg.

    It is to mimic the bent leg.

    Args:
      value: A list of four values.
    """
    self._swing_offset = value

  def set_extension_offset(self, value):
    """Set the extension offset of each leg.

    It is to mimic the bent leg.

    Args:
      value: A list of four values.
    """
    self._extension_offset = value

  def set_desired_pitch(self, value):
    """Set the desired pitch of the base, which is a user input.

    Args:
      value: A scalar.
    """
    self.desired_pitch = value



"""Tests for pybullet_envs.minitaur.robots.mini_cheetah.

blaze test -c opt
//robotics/reinforcement_learning/minitaur/robots:mini_cheetah_test
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import numpy as np
from pybullet_envs.minitaur.envs import bullet_client
from pybullet_envs.minitaur.robots import mini_cheetah
from google3.testing.pybase import googletest

PI = math.pi
NUM_STEPS = 500
TIME_STEP = 0.002
INIT_MOTOR_ANGLES = [0, -0.6, 1.4] * 4


class MiniCheetahTest(googletest.TestCase):

  def test_init(self):
    pybullet_client = bullet_client.BulletClient()
    pybullet_client.enable_cns()
    robot = mini_cheetah.MiniCheetah(
        pybullet_client=pybullet_client, time_step=TIME_STEP, on_rack=True)
    self.assertIsNotNone(robot)

  def test_static_pose_on_rack(self):
    pybullet_client = bullet_client.BulletClient()
    pybullet_client.enable_cns()
    pybullet_client.resetSimulation()
    pybullet_client.setPhysicsEngineParameter(numSolverIterations=60)
    pybullet_client.setTimeStep(TIME_STEP)
    pybullet_client.setGravity(0, 0, -10)

    robot = (
        mini_cheetah.MiniCheetah(
            pybullet_client=pybullet_client,
            action_repeat=5,
            time_step=0.002,
            on_rack=True))
    robot.Reset(
        reload_urdf=False,
        default_motor_angles=INIT_MOTOR_ANGLES,
        reset_time=0.5)
    for _ in range(NUM_STEPS):
      robot.Step(INIT_MOTOR_ANGLES)
      motor_angles = robot.GetMotorAngles()
      np.testing.assert_array_almost_equal(
          motor_angles, INIT_MOTOR_ANGLES, decimal=2)


if __name__ == '__main__':
  googletest.main()



import pybullet as p
import time
import math
from datetime import datetime
import pybullet_data

clid = p.connect(p.SHARED_MEMORY)
if (clid < 0):
  p.connect(p.GUI)
  #p.connect(p.SHARED_MEMORY_GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())

p.loadURDF("plane.urdf", [0, 0, -0.3])
kukaId = p.loadURDF("kuka_iiwa/model.urdf", [0, 0, 0])
p.resetBasePositionAndOrientation(kukaId, [0, 0, 0], [0, 0, 0, 1])
kukaEndEffectorIndex = 6
numJoints = p.getNumJoints(kukaId)
if (numJoints != 7):
  exit()

#lower limits for null space
ll = [-.967, -2, -2.96, 0.19, -2.96, -2.09, -3.05]
#upper limits for null space
ul = [.967, 2, 2.96, 2.29, 2.96, 2.09, 3.05]
#joint ranges for null space
jr = [5.8, 4, 5.8, 4, 5.8, 4, 6]
#restposes for null space
rp = [0, 0, 0, 0.5 * math.pi, 0, -math.pi * 0.5 * 0.66, 0]
#joint damping coefficents
jd = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

for i in range(numJoints):
  p.resetJointState(kukaId, i, rp[i])

p.setGravity(0, 0, 0)
t = 0.
prevPose = [0, 0, 0]
prevPose1 = [0, 0, 0]
hasPrevPose = 0
useNullSpace = 1

useOrientation = 1
#If we set useSimulation=0, it sets the arm pose to be the IK result directly without using dynamic control.
#This can be used to test the IK result accuracy.
useSimulation = 1
useRealTimeSimulation = 0
ikSolver = 0
p.setRealTimeSimulation(useRealTimeSimulation)
#trailDuration is duration (in seconds) after debug lines will be removed automatically
#use 0 for no-removal
trailDuration = 15

i=0
while 1:
  i+=1
  #p.getCameraImage(320,
  #                 200,
  #                 flags=p.ER_SEGMENTATION_MASK_OBJECT_AND_LINKINDEX,
  #                 renderer=p.ER_BULLET_HARDWARE_OPENGL)
  if (useRealTimeSimulation):
    dt = datetime.now()
    t = (dt.second / 60.) * 2. * math.pi
  else:
    t = t + 0.01

  if (useSimulation and useRealTimeSimulation == 0):
    p.stepSimulation()

  for i in range(1):
    pos = [-0.4, 0.2 * math.cos(t), 0. + 0.2 * math.sin(t)]
    #end effector points down, not up (in case useOrientation==1)
    orn = p.getQuaternionFromEuler([0, -math.pi, 0])

    if (useNullSpace == 1):
      if (useOrientation == 1):
        jointPoses = p.calculateInverseKinematics(kukaId, kukaEndEffectorIndex, pos, orn, ll, ul,
                                                  jr, rp)
      else:
        jointPoses = p.calculateInverseKinematics(kukaId,
                                                  kukaEndEffectorIndex,
                                                  pos,
                                                  lowerLimits=ll,
                                                  upperLimits=ul,
                                                  jointRanges=jr,
                                                  restPoses=rp)
    else:
      if (useOrientation == 1):
        jointPoses = p.calculateInverseKinematics(kukaId,
                                                  kukaEndEffectorIndex,
                                                  pos,
                                                  orn,
                                                  jointDamping=jd,
                                                  solver=ikSolver,
                                                  maxNumIterations=100,
                                                  residualThreshold=.01)
      else:
        jointPoses = p.calculateInverseKinematics(kukaId,
                                                  kukaEndEffectorIndex,
                                                  pos,
                                                  solver=ikSolver)

    if (useSimulation):
      for i in range(numJoints):
        p.setJointMotorControl2(bodyIndex=kukaId,
                                jointIndex=i,
                                controlMode=p.POSITION_CONTROL,
                                targetPosition=jointPoses[i],
                                targetVelocity=0,
                                force=500,
                                positionGain=0.03,
                                velocityGain=1)
    else:
      #reset the joint state (ignoring all dynamics, not recommended to use during simulation)
      for i in range(numJoints):
        p.resetJointState(kukaId, i, jointPoses[i])

  ls = p.getLinkState(kukaId, kukaEndEffectorIndex)
  if (hasPrevPose):
    p.addUserDebugLine(prevPose, pos, [0, 0, 0.3], 1, trailDuration)
    p.addUserDebugLine(prevPose1, ls[4], [1, 0, 0], 1, trailDuration)
  prevPose = pos
  prevPose1 = ls[4]
  hasPrevPose = 1
p.disconnect()



"""A wrapped MinitaurGymEnv with a built-in controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from gym import spaces
import numpy as np
import gin
from pybullet_envs.minitaur.agents.trajectory_generator import tg_simple
from pybullet_envs.minitaur.envs_v2.utilities import robot_pose_utils
from pybullet_envs.minitaur.robots.utilities import action_filter

_DELTA_TIME_LOWER_BOUND = 0.0
_DELTA_TIME_UPPER_BOUND = 3

_GAIT_PHASE_MAP = {
    "walk": [0, 0.25, 0.5, 0.75],
    "trot": [0, 0.5, 0.5, 0],
    "bound": [0, 0.5, 0, 0.5],
    "pace": [0, 0, 0.5, 0.5],
    "pronk": [0, 0, 0, 0]
}


@gin.configurable
class PmtgWrapperEnv(object):
  """A wrapped LocomotionGymEnv with a built-in trajectory generator."""

  def __init__(self,
               gym_env,
               intensity_upper_bound=1.5,
               min_delta_time=_DELTA_TIME_LOWER_BOUND,
               max_delta_time=_DELTA_TIME_UPPER_BOUND,
               integrator_coupling_mode="all coupled",
               walk_height_coupling_mode="all coupled",
               variable_swing_stance_ratio=True,
               swing_stance_ratio=1.0,
               residual_range=0.15,
               init_leg_phase_offsets=None,
               init_gait=None,
               default_walk_height=0,
               action_filter_enable=True,
               action_filter_order=1,
               action_filter_low_cut=0,
               action_filter_high_cut=3.0,
               action_filter_initialize=False,
               leg_pose_class=None):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      intensity_upper_bound: The upper bound for intensity of the trajectory
        generator. It can be used to limit the leg movement.
      min_delta_time: Lower limit for the time in seconds that the trajectory
        generator can be moved forward at each simulation step. The effective
        frequency of the gait is based on the delta time multiplied by the
        internal frequency of the trajectory generator.
      max_delta_time: Upper limit for the time in seconds that the trajectory
        generator can be moved forward at each simulation step.
      integrator_coupling_mode: How the legs should be coupled for integrators.
      walk_height_coupling_mode: The same coupling mode used for walking walking
        heights for the legs.
      variable_swing_stance_ratio: A boolean to indicate if the swing stance
        ratio can change per time step or not.
      swing_stance_ratio: Time taken by swing phase vs stance phase. This is
        only relevant if variable_swing_stance_ratio is False.
      residual_range: The upper limit for the residual actions that adds to the
        leg motion. It is 0.15 by default, can be increased for more flexibility
        or decreased to only to use the trajectory generator's motion.
      init_leg_phase_offsets: The initial phases of the legs. A list of 4
        variables within [0,1). The order is front-left, rear-left, front-right
        and rear-right.
      init_gait: The initial gait that sets the starting phase difference
        between the legs. Overrides the arg init_phase_offsets. Has to be
        "walk", "trot", "bound" or "pronk". Used in vizier search.
      default_walk_height: Offset for the extension of the legs for the robot.
        Applied to the legs at every time step. Implicitly affects the walking
        and standing height of the policy. Zero by default. Units is in
        extension space (can be considered in radiant since it is a linear
        transformation to motor angles based on the robot's geometry).
      action_filter_enable: Use a butterworth filter for the output of the PMTG
        actions (before conversion to leg swing-extend model). It forces
        smoother behaviors depending on the parameters used.
      action_filter_order: The order for the action_filter (1 by default).
      action_filter_low_cut: The cut for the lower frequencies (0 by default).
      action_filter_high_cut: The cut for the higher frequencies (3 by default).
      action_filter_initialize: If the action filter should be initialized when
        the first action is taken. If enabled, the filter does not affect action
        value the first time it is called and fills the history with that value.
      leg_pose_class: A class providing a convert_leg_pose_to_motor_angle
        instance method or None. If None, robot_pose_utils is used.

    Raises:
      ValueError if the controller does not implement get_action and
      get_observation.

    """
    self._gym_env = gym_env
    self._num_actions = gym_env.robot.num_motors
    self._residual_range = residual_range
    self._min_delta_time = min_delta_time
    self._max_delta_time = max_delta_time
    self._leg_pose_util = leg_pose_class() if leg_pose_class else None
    # If not specified, default leg phase offsets to walking.
    if init_gait:
      if init_gait in _GAIT_PHASE_MAP:
        init_leg_phase_offsets = _GAIT_PHASE_MAP[init_gait]
      else:
        raise ValueError("init_gait is not one of the defined gaits.")
    else:
      init_leg_phase_offsets = init_leg_phase_offsets or [0, 0.25, 0.5, 0.75]
    # Create the Trajectory Generator based on the parameters.
    self._trajectory_generator = tg_simple.TgSimple(
        intensity_upper_bound=intensity_upper_bound,
        integrator_coupling_mode=integrator_coupling_mode,
        walk_height_coupling_mode=walk_height_coupling_mode,
        variable_swing_stance_ratio=variable_swing_stance_ratio,
        swing_stance_ratio=swing_stance_ratio,
        init_leg_phase_offsets=init_leg_phase_offsets)

    action_dim = self._extend_action_space()
    self._extend_obs_space()

    self._default_walk_height = default_walk_height
    self._action_filter_enable = action_filter_enable
    if self._action_filter_enable:
      self._action_filter_initialize = action_filter_initialize
      self._action_filter_order = action_filter_order
      self._action_filter_low_cut = action_filter_low_cut
      self._action_filter_high_cut = action_filter_high_cut
      self._action_filter = self._build_action_filter(action_dim)

  def _extend_obs_space(self):
    """Extend observation space to include pmtg phase variables."""
    # Set the observation space and boundaries.
    lower_bound, upper_bound = self._get_observation_bounds()
    if hasattr(self._gym_env.observation_space, "spaces"):
      new_obs_space = spaces.Box(np.array(lower_bound), np.array(upper_bound))
      self.observation_space.spaces.update({"pmtg_phase": new_obs_space})
    else:
      lower_bound = np.append(self._gym_env.observation_space.low, lower_bound)
      upper_bound = np.append(self._gym_env.observation_space.high, upper_bound)
      self.observation_space = spaces.Box(
          np.array(lower_bound), np.array(upper_bound), dtype=np.float32)

  def _extend_action_space(self):
    """Extend the action space to include pmtg parameters."""
    # Add the action boundaries for delta time, one per integrator.
    action_low = [-self._residual_range] * self._num_actions
    action_high = [self._residual_range] * self._num_actions
    action_low = np.append(action_low, [self._min_delta_time] *
                           self._trajectory_generator.num_integrators)
    action_high = np.append(action_high, [self._max_delta_time] *
                            self._trajectory_generator.num_integrators)

    # Add the action boundaries for parameters of the trajectory generator.
    l_bound, u_bound = self._trajectory_generator.get_parameter_bounds()
    action_low = np.append(action_low, l_bound)
    action_high = np.append(action_high, u_bound)
    self.action_space = spaces.Box(
        np.array(action_low), np.array(action_high), dtype=np.float32)
    return len(action_high)

  def __getattr__(self, attrb):
    return getattr(self._gym_env, attrb)

  def _modify_observation(self, observation):
    if isinstance(observation, dict):
      observation["pmtg_phase"] = self._trajectory_generator.get_state()
      return observation
    else:
      return np.append(observation, self._trajectory_generator.get_state())

  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    """Resets the environment as well as the trajectory generator(s).

    Args:
      initial_motor_angles: Unused for PMTG. Instead, it sets the legs to a pose
        with the neutral action for the trajectory generator.
      reset_duration: Float. The time (in seconds) needed to rotate all motors
        to the desired initial values.

    Returns:
      A numpy array contains the initial observation after reset.
    """
    del initial_motor_angles
    if self._action_filter_enable:
      self._reset_action_filter()
    self._last_real_time = 0
    self._num_step = 0
    self._target_speed_coef = 0.0
    if self._trajectory_generator:
      self._trajectory_generator.reset()
    if self._leg_pose_util:
      initial_motor_angles = self._leg_pose_util.convert_leg_pose_to_motor_angles(
          [0, 0, 0] * 4)
    else:
      initial_motor_angles = robot_pose_utils.convert_leg_pose_to_motor_angles(
          self._gym_env.robot_class, [0, 0, 0] * 4)
    observation = self._gym_env.reset(initial_motor_angles, reset_duration)
    return self._modify_observation(observation)

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the modified observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.

    """

    if action is None:
      raise ValueError("Action cannot be None")

    if self._action_filter_enable:
      action = self._filter_action(action)

    time = self._gym_env.get_time_since_reset()

    action_residual = action[0:self._num_actions]
    # Add the default walking height offset to extension.
    dimensions = len(action_residual) // 4
    action_residual[dimensions - 1::dimensions] += self._default_walk_height
    # Calculate trajectory generator's output based on the rest of the actions.
    delta_real_time = time - self._last_real_time
    self._last_real_time = time
    action_tg = self._trajectory_generator.get_actions(
        delta_real_time, action[self._num_actions:])
    # If the residuals have a larger dimension, extend trajectory generator's
    # actions to include abduction motors.
    if len(action_tg) == 8 and len(action_residual) == 12:
      for i in [0, 3, 6, 9]:
        action_tg.insert(i, 0)
    # Add TG actions with residual actions (in swing - extend space).
    action_total = [a + b for a, b in zip(action_tg, action_residual)]
    # Convert them to motor space based on the robot-specific conversions.
    if self._leg_pose_util:
      action_motor_space = self._leg_pose_util.convert_leg_pose_to_motor_angles(
          action_total)
    else:
      action_motor_space = robot_pose_utils.convert_leg_pose_to_motor_angles(
          self._gym_env.robot_class, action_total)
    original_observation, reward, done, _ = self._gym_env.step(
        action_motor_space)

    return self._modify_observation(original_observation), np.float32(
        reward), done, _

  def _get_observation_bounds(self):
    """Get the bounds of the observation added from the trajectory generator.

    Returns:
      lower_bounds: Lower bounds for observations.
      upper_bounds: Upper bounds for observations.
    """
    lower_bounds = self._trajectory_generator.get_state_lower_bounds()
    upper_bounds = self._trajectory_generator.get_state_upper_bounds()
    return lower_bounds, upper_bounds

  def _build_action_filter(self, num_joints):
    order = self._action_filter_order
    low_cut = self._action_filter_low_cut
    high_cut = self._action_filter_high_cut
    sampling_rate = 1 / (0.01)
    a_filter = action_filter.ActionFilterButter([low_cut], [high_cut],
                                                sampling_rate, order,
                                                num_joints)
    return a_filter

  def _reset_action_filter(self):
    self._action_filter.reset()
    self._action_filter_empty = True
    return

  def _filter_action(self, action):
    if self._action_filter_empty and self._action_filter_initialize:
      # If initialize is selected and it is the first time filter is called,
      # fill the buffer with that action so that it starts from that value
      # instead of zero(s).
      init_action = np.array(action).reshape(len(action), 1)
      self._action_filter.reset(init_action)
      self._action_filter_empty = False
    filtered_action = self._action_filter.filter(np.array(action))
    return filtered_action



"""A config file for parameters and their ranges in dynamics randomization."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

PARAM_RANGE = {
    # The following ranges are in percentage. e.g. 0.8 means 80%.
    "mass": [0.8, 1.2],
    "inertia": [0.5, 1.5],
    "motor strength": [0.8, 1.2],
    # The following ranges are the physical values, in SI unit.
    "motor friction": [0, 0.05],  # Viscous damping (Nm s/rad).
    "control step": [0.003, 0.02],  # Time inteval (s).
    "latency": [0.0, 0.04],  # Time inteval (s).
    "lateral friction": [0.5, 1.25],  # Friction coefficient (dimensionless).
    "battery": [14.0, 16.8],  # Voltage (V).
    "joint friction": [0, 0.05],  # Coulomb friction torque (Nm).
}


def all_params():
  """Randomize all the physical parameters."""
  return PARAM_RANGE
